import sys
sys.path.append('..')

from neml import interpolate, solvers, models, elasticity, ri_flow, hardening, surfaces, visco_flow, general_flow, creep
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonMatModel(object):
  """
    Tests that could be applied to all material models
  """
  def test_tangent_proportional_strain(self):
    t_n = 0.0
    strain_n = np.zeros((6,))
    stress_n = np.zeros((6,))
    hist_n = self.model.init_store()

    u_n = 0.0
    p_n = 0.0

    for i,m in enumerate(np.linspace(0,1,self.nsteps)):
      t_np1 = self.tfinal * m
      strain_np1 = self.efinal * m
      
      stress_np1, hist_np1, A_np1, u_np1, p_np1 = self.model.update_sd(
          strain_np1, strain_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)
      dfn = lambda e: self.model.update_sd(e,
          strain_n, self.T, self.T, t_np1, t_n, stress_n, hist_n, u_n, p_n)[0]
      num_A = differentiate(dfn, strain_np1, eps = 1.0e-9)
      
      if not np.allclose(num_A, A_np1, rtol = 1e-3, atol = 1e-1):
        print(A_np1)
        print(num_A)

      self.assertTrue(np.allclose(num_A, A_np1, rtol = 1.0e-3, atol = 1.0e-1))
      
      strain_n = strain_np1
      stress_n = stress_np1
      hist_n = hist_np1
      u_n = u_np1
      p_n = p_np1

  def test_elastic_proportional_strain(self):
    t_n = 0.0
    strain_n = np.zeros((6,))
    stress_n = np.zeros((6,))
    hist_n = self.model.init_store()

    u_n = 0.0
    p_n = 0.0

    for i,m in enumerate(np.linspace(0,1,self.nsteps)):
      t_np1 = self.tfinal * m
      strain_np1 = self.efinal * m
      
      stress_np1, hist_np1, A_np1, u_np1, p_np1 = self.model.update_sd(
          strain_np1, strain_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)
      
      estrain = self.model.elastic_strains(stress_np1, self.T, hist_np1)
      S = self.elastic.S(self.T)
      estrain_num = np.dot(S, stress_np1)
      self.assertTrue(np.allclose(estrain, estrain_num))
      
      strain_n = strain_np1
      stress_n = stress_np1
      hist_n = hist_np1
      u_n = u_np1
      p_n = p_np1

class TestLinearElastic(CommonMatModel, unittest.TestCase):
  """
    Linear elasticity, as a benchmark
  """
  def setUp(self):
    self.hist0 = np.array([])
    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")
    self.model = models.SmallStrainElasticity(self.elastic)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

class CommonJacobian(object):
  """
    Common jacobian tests
  """
  def gen_strain(self):
    return np.array([0.01,-0.01,0.02,0.03,-0.005,0.01])

  def gen_T(self):
    return 500.0

  def gen_time(self):
    return 1.25

  def gen_start_strain(self):
    return np.zeros((6,))

  def gen_start_time(self):
    return 0.0

  def gen_start_stress(self):
    return np.zeros((6,))

  def test_jacobian(self):
    e_np1 = self.gen_strain()
    T_np1 = self.gen_T()
    h_n = self.gen_hist()
    t_np1 = self.gen_time()

    ts = self.model.make_trial_state(e_np1, self.gen_start_strain(), T_np1, T_np1,
        t_np1, self.gen_start_time(), self.gen_start_stress(), h_n)

    x = self.gen_x()

    R, J = self.model.RJ(x, ts)
    
    dfn = lambda y: self.model.RJ(y, ts)[0]
    nJ = differentiate(dfn, x)
   
    self.assertTrue(np.allclose(J, nJ, rtol = 1.0e-3, atol = 1e-1))

class TestPerfectPlasticity(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test J2 perfect plasticity
  """
  def setUp(self):
    self.hist0 = np.zeros((0,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    self.model = models.SmallStrainPerfectPlasticity(self.elastic, surface, 
        self.s0)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def test_properties(self):
    self.assertTrue(np.isclose(self.model.ys(self.T), self.s0))

  def gen_hist(self):
    return np.zeros((0,))

  def gen_x(self):
    return np.array([50.0,100.0,-25.0,150.0,200.0,-50.0] + list([0.25]))

class TestRIAPlasticityCombinedLinearLinear(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test combined isotropic/kinematic hardening with linear rules
  """
  def setUp(self):
    self.hist0 = np.zeros((13,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.Kp = 1000.0

    self.H = 1000.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    kin = hardening.LinearKinematicHardeningRule(self.H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model = models.SmallStrainRateIndependentPlasticity(self.elastic, flow)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def gen_hist(self):
    h = np.array(range(1,14)) / 15.0
    h[:6] = make_dev(h[:6])
    h[7:] = make_dev(h[7:])
    return h

  def gen_x(self):
    return np.array(list(self.gen_hist()) + list([0.25]))

class TestRIAPlasticityJ2Linear(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test the rate-independent plasticity algorithm with a linearly
    isotropically hardening yield surface.
  """
  def setUp(self):
    self.hist0 = np.zeros((7,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.Kp = self.E / 100

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model = models.SmallStrainRateIndependentPlasticity(self.elastic, flow)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def gen_hist(self):
    return np.array(range(1,8)) / 8.0

  def gen_x(self):
    return np.array(range(1,9)) / 9.0


class TestRIAPlasticityJ2Voce(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test the rate-independent plasticity algorithm with a Voce
    isotropically hardening yield surface.
  """
  def setUp(self):
    self.hist0 = np.zeros((7,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.R = 150.0
    self.d = 10.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.VoceIsotropicHardeningRule(self.s0, self.R, self.d)
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    self.model = models.SmallStrainRateIndependentPlasticity(self.elastic, flow)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def gen_hist(self):
    return np.array(range(1,8)) / 8.0

  def gen_x(self):
    return np.array(range(1,9)) / 9.0


class TestRIChabocheLinear(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test Chaboche with linear isotropic hardening
  """
  def setUp(self):
    self.hist0 = np.zeros((13,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.Kp = 1000.0

    self.n = 2
    self.cs = [10.0, 2.0]
    self.rs = [5.0, 1.0]
    self.gmodels = [hardening.ConstantGamma(g) for g in self.rs]
    self.As = [0.0] * self.n
    self.ns = [1.0] * self.n

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(self.s0, self.Kp)
    hmodel = hardening.Chaboche(iso, self.cs, self.gmodels, 
        self.As, self.ns)

    flow = ri_flow.RateIndependentNonAssociativeHardening(surface, hmodel)

    self.model = models.SmallStrainRateIndependentPlasticity(self.elastic, flow)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 30

  def gen_hist(self):
    h = np.array(range(1,6+1+6*self.n+1)) / (8*self.n)
    h[:6] = make_dev(h[:6])
    for i in range(self.n):
      h[7+i*6:7+(i+1)*6] = make_dev(h[7+i*6:7+(i+1)*6])
    return h

  def gen_x(self):
    return np.array(list(self.gen_hist()) + [0.1])

# Something funny with the Jacobian here
class TestCreepPlasticityJ2LinearPowerLaw(unittest.TestCase, CommonMatModel):
  """
    Test the combined creep/plasticity algorithm with J2 plasticity with
    isotropic hardening and power law creep 
  """
  def setUp(self):
    self.hist0 = np.zeros((13,))

    self.A = 1.85e-10
    self.n = 2.5

    self.smodel = creep.PowerLawCreep(self.A, self.n)
    self.cmodel = creep.J2CreepModel(self.smodel)

    self.E = 150000.0
    self.nu = 0.3
    self.sY = 200.0
    self.H = self.E / 50.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs", 
        self.nu, "poissons")

    self.surface = surfaces.IsoJ2()
    self.iso = hardening.LinearIsotropicHardeningRule(self.sY, self.H)
    self.flow = ri_flow.RateIndependentAssociativeFlow(self.surface, self.iso)

    self.pmodel = models.SmallStrainRateIndependentPlasticity(self.elastic, 
        self.flow)

    self.model = models.SmallStrainCreepPlasticity(self.elastic, self.pmodel,
        self.cmodel)

    self.efinal = np.array([0.1,-0.05,0.02,-0.03,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def gen_hist(self):
    return np.array(range(1,7) + [1.0] + range(1,7)) / 7.0

  def gen_x(self):
    return np.array(range(1,7)) / 7.0

  def gen_start_stress(self):
    return np.zeros((6,)) + 100.0

  def gen_start_strain(self):
    return np.zeros((6,)) + 0.01

class TestCreepPlasticityPerfect(unittest.TestCase, CommonMatModel):
  """
    Test the combined creep/plasticity algorithm with J2 plasticity with
    perfect plasticity
  """
  def setUp(self):
    self.hist0 = np.zeros((6,))

    self.A = 1.85e-10
    self.n = 2.5

    self.smodel = creep.PowerLawCreep(self.A, self.n)
    self.cmodel = creep.J2CreepModel(self.smodel)

    self.E = 150000.0
    self.nu = 0.3
    self.sY = 200.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E, "youngs", 
        self.nu, "poissons")
    self.surface = surfaces.IsoJ2()

    self.pmodel = models.SmallStrainPerfectPlasticity(self.elastic, 
        self.surface, self.sY)

    self.model = models.SmallStrainCreepPlasticity(self.elastic, 
        self.pmodel, self.cmodel)

    self.efinal = np.array([0.1,-0.05,0.02,-0.05,0.1,-0.15])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 10

  def gen_hist(self):
    return np.array(range(1,7) + [1.0] + range(1,7)) / 7.0

  def gen_x(self):
    return np.array(range(1,7)) / 7.0

  def gen_start_stress(self):
    return np.zeros((6,)) + 100.0

  def gen_start_strain(self):
    return np.zeros((6,)) + 0.01

class TestDirectIntegrateChaboche(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test Chaboche's VP model with our new direct integrator
  """
  def setUp(self):
    n = 20.0
    eta = 108.0
    sY = 89.0

    Q = 165.0
    b = 12.0
    
    self.m = 3

    C1 = 80.0e3
    C2 = 14.02e3
    C3 = 3.333e3

    y1 = 0.9e3
    y2 = 1.5e3
    y3 = 1.0

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(sY, Q, b)
    cs = [C1, C2, C3]
    gs = [y1, y2, y3]
    As = [0.0, 0.0, 0.0]
    ns = [1.0, 1.0, 1.0]
    gmodels = [hardening.ConstantGamma(g) for g in gs]
    hmodel = hardening.Chaboche(iso, cs, gmodels, As, ns)

    fluidity = visco_flow.ConstantFluidity(eta)

    self.hist0 = np.zeros((19,))
    self.T = 300.0

    vmodel = visco_flow.ChabocheFlowRule(surface, hmodel, fluidity, n)

    E = 92000.0
    nu = 0.3

    mu = E/(2*(1+nu))
    K = E/(3*(1-2*nu))

    self.elastic = elasticity.IsotropicLinearElasticModel(mu,
        "shear", K, "bulk")

    flow = general_flow.TVPFlowRule(self.elastic, vmodel)

    self.model = models.GeneralIntegrator(self.elastic, flow)

    self.efinal = np.array([0.05,0,0,0.02,0,-0.01])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 100

  def gen_hist(self):
    h = np.array([0.05,20,-30,40.0,5.0,2.0,40.0,-10,-20,10,50,30,10,-50,60,70,15,-15,10])
    h[1:7] = make_dev(h[1:7])
    h[7:13] = make_dev(h[7:13])
    h[13:19] = make_dev(h[13:19])
    return h

  def gen_x(self):
    x = [100.0,150.0,-300.0,-10.0,50.0,100.0] + list(self.gen_hist()*1.1)
    return np.array(x)

class TestPerzynaJ2Voce(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Perzyna associated viscoplasticity w/ voce kinematic hardening
  """
  def setUp(self):
    self.hist0 = np.zeros((7,))

    self.E = 92000.0
    self.nu = 0.3

    self.mu = self.E/(2*(1+self.nu))
    self.K = self.E/(3*(1-2*self.nu))

    self.s0 = 180.0
    self.R = 150.0
    self.d = 10.0

    self.n = 2.0
    self.eta = 200.0

    self.elastic = elasticity.IsotropicLinearElasticModel(self.mu,
        "shear", self.K, "bulk")

    surface = surfaces.IsoJ2()
    hrule = hardening.VoceIsotropicHardeningRule(self.s0, self.R, self.d)
    g = visco_flow.GPowerLaw(self.n, self.eta)

    vmodel = visco_flow.PerzynaFlowRule(surface, hrule, g)

    flow = general_flow.TVPFlowRule(self.elastic, vmodel)
    self.model = models.GeneralIntegrator(self.elastic, flow)

    self.efinal = np.array([0.05,0,0,0.02,0,-0.01])
    self.tfinal = 10.0
    self.T = 300.0
    self.nsteps = 100

  def gen_hist(self):
    return np.array(range(1,8)) / 8.0

  def gen_x(self):
    return np.array(range(1,8)) / 8.0

class TestYaguchi(unittest.TestCase, CommonMatModel, CommonJacobian):
  """
    Test Cheboche's VP model with our new direct integrator
  """
  def setUp(self):
    vmodel = visco_flow.YaguchiGr91FlowRule()

    E_poly = np.array([-3.077e-1, 3.003e2, 1.269e5])
    nu = 0.3

    mu_poly = list(E_poly/(2*(1+nu)))
    K_poly = list(E_poly/(3*(1-2*nu)))

    shear = interpolate.PolynomialInterpolate(mu_poly)
    bulk = interpolate.PolynomialInterpolate(K_poly)
    self.elastic = elasticity.IsotropicLinearElasticModel(shear, "shear", 
        bulk, "bulk")

    flow = general_flow.TVPFlowRule(self.elastic, vmodel)

    self.model = models.GeneralIntegrator(self.elastic, flow)

    self.efinal = np.array([0.05,0,0,0.02,0,-0.01])
    self.tfinal = 10.0
    self.T = 550.0
    self.nsteps = 100

  def gen_hist(self):
    h = np.array([20,-30,40.0,5.0,2.0,40.0,-10,-20,10,50,30,10,10.0,5.0])
    h[:6] = make_dev(h[:6])
    h[6:12] = make_dev(h[6:12])
    return h

  def gen_x(self):
    x = [100.0,150.0,-300.0,-10.0,50.0,100.0] + list(self.gen_hist()*1.1)
    return np.array(x)

class TestKMSwitch(unittest.TestCase, CommonMatModel):
  """
    Test the model with a switch
  """
  def setUp(self):
    # Fully-defined perfectly plastic model
    Epoly = [-78.2759, 236951.0]
    nu = 0.3
    A = -9.6187
    B = -1.4819
    C = -5.0486
    g0 = 0.3708

    b = 0.248 * 1.0e-6
    kboltz = 1.38064e-23 * 1000.0
    eps0 = 1.0e10

    # Temperature range over which to consider (K)
    Tmin = 550.0
    Tmax = 950.0
    Trange = np.linspace(Tmin, Tmax)

    # Elastic
    E_m = interpolate.PolynomialInterpolate(Epoly)
    nu_m = interpolate.ConstantInterpolate(nu)
    elastic_m = elasticity.IsotropicLinearElasticModel(E_m, "youngs",
        nu_m, "poissons")
    self.elastic = elastic_m

    # Rate sensitivity interpolates values
    mu_values = np.array([elastic_m.G(T) for T in Trange])
    n_values = -mu_values*b**3.0 / (kboltz * Trange * A)
    eta_values = np.exp(B) * eps0 ** (kboltz * Trange * A / (mu_values * b**3.0)) * mu_values

    # Rate independent interpolate values
    flow_stress = mu_values * np.exp(C)
    
    # Common objects
    surface = surfaces.IsoKinJ2()
    hmodulus = interpolate.PolynomialInterpolate([-10.0, 12000.0])

    # Setup visco model
    n_interp = interpolate.PiecewiseLinearInterpolate(list(Trange), list(n_values))
    eta_interp = interpolate.PiecewiseLinearInterpolate(list(Trange), list(eta_values))
    eta_m = visco_flow.ConstantFluidity(eta_interp)

    iso_rd = hardening.LinearIsotropicHardeningRule(
        interpolate.ConstantInterpolate(0.0),
        hmodulus)
    hard_rd = hardening.Chaboche(iso_rd,
        [interpolate.ConstantInterpolate(0.0)], 
        [hardening.ConstantGamma(interpolate.ConstantInterpolate(0.0))],
        [interpolate.ConstantInterpolate(0.0)],
        [interpolate.ConstantInterpolate(1.0)])

    visco_rd = visco_flow.ChabocheFlowRule(surface, hard_rd, eta_m, n_interp) 
    general_rd = general_flow.TVPFlowRule(elastic_m, visco_rd)

    rate_dependent = models.GeneralIntegrator(elastic_m, general_rd)

    # Setup rate independent
    sy_interp = interpolate.PiecewiseLinearInterpolate(list(Trange), list(flow_stress))
    iso_ri = hardening.LinearIsotropicHardeningRule(sy_interp, hmodulus)
    hard_ri = hardening.Chaboche(iso_ri,
        [interpolate.ConstantInterpolate(0.0)], 
        [hardening.ConstantGamma(interpolate.ConstantInterpolate(0.0))],
        [interpolate.ConstantInterpolate(0.0)],
        [interpolate.ConstantInterpolate(1.0)])
    flow_ri = ri_flow.RateIndependentNonAssociativeHardening(surface, hard_ri)
    
    rate_independent = models.SmallStrainRateIndependentPlasticity(elastic_m,
        flow_ri)

    # Combined model
    self.model = models.KMRegimeModel(elastic_m, [rate_independent, rate_dependent],
        [g0], kboltz, b, eps0)

    self.efinal = np.array([0.05,0,0,0.02,0,-0.01])
    self.tfinal = 10.0
    self.T = 700.0
    self.nsteps = 100

  def gen_hist(self):
    h = np.array([40.0,20,-30,40.0,5.0,2.0,40.0])
    h[:6] = make_dev(h[:6])
    return h



from neml.math import tensors, rotations
from neml.cp import crystallography

import numpy as np
import numpy.linalg as la

from matplotlib.patches import Polygon
import matplotlib.pyplot as plt

limit_lambert_equal_area = np.sqrt(2)
limit_stereographic = 1.0

triangle_limits = {"cubic": (0,0)}

def project_stereographic(v):
  """
    Stereographic projection of the given vector into a numpy array

    Parameters:
      v:        input vector
  """
  return np.array([v[0]/(1.0+v[2]), v[1]/(1.0+v[2])])

def project_lambert_equal_area(v):
  """
    Lambert equal area projection of the given vector into a numpy array

    Parameters:
      v:        input vector
  """
  return np.array([np.sqrt(2/(1+v[2]))*v[0]/np.sqrt(2),
    np.sqrt(2/(1+v[2]))*v[0]/np.sqrt(2)])

def inverse_project_stereographic(pt):
  """
    Inverse stereographic projection

    Parameters:
      pt:            point, as numpy array
  """
  X = pt[0]
  Y = pt[1]
  bot = 1.0 + X**2.0 + Y**2.0
  return tensors.Vector([2.0*X/bot,2.0*Y/bot,(X**2.0+Y**2.0-1.0)/bot])

def inverse_project_lambert_equal_area(pt):
  """
    Inverse Lambert projections

    Parameters:
      pt:            point, as a numpy array
  """
  X = pt[0]
  Y = pt[1]
  f = np.sqrt(1.0-(X**2.0+Y**2.0)/4)
  return tensors.Vector([f*X,f*Y,-1.0+(X**2.0+Y**2.0)/2])

def cart2pol(x):
  """
    Convert a cartesian point into polar coordinates

    Parameters:
      x:        point to convert to polar
  """
  return np.array([np.arctan2(x[1], x[0]), la.norm(x)])

def pole_figure_discrete(orientations, pole, lattice,  
    projection = "stereographic", 
    sample_symmetry = crystallography.symmetry_rotations("222"),
    x = tensors.Vector([1.0,0,0]), y = tensors.Vector([0,1.0,0]),
    axis_labels = ["X", "Y"]):
  """
    Plot a standard pole figure given a collection of discrete points.

    Parameters:
      orientations:      list of orientations
      pole:              pole as a integer list
      lattice:           crystal lattice class, including crystal symmetry

    Keyword Args:
      projection:        which projection to use
      sample_symmetry:   what sample symmetry to apply to the pole figure,
                         defaults to orthorhombic
      x:                 x direction for plot, defaults to [1,0,0]
      y:                 y direction for plot, defaults to [0,1,0]
      axis_labels:       axis labels to include on the figure
  """
  # Get the rotation from the standard coordinates to the given x and y
  srot = rotations.Orientation(x,y)
  
  # Get all equivalent poles
  p = lattice.miller2cart_direction(pole)
  eq_poles = lattice.equivalent_vectors(p)
  eq_poles = [pp.normalize() for pp in eq_poles]

  # Apply sample symmetric
  eq_poles = [op.apply(p) for p in eq_poles for op in sample_symmetry]
  
  # Get the points on the sphere
  pts = [srot.apply(o.inverse().apply(pp)) for pp in eq_poles for o in orientations]

  # Get rid of those that are in the lower hemisphere
  pts = [p for p in pts if p[2] > 0.0]

  # Project
  if projection == "stereographic":
    pop = project_stereographic
    lim = limit_stereographic
  elif projection == "lambert":
    pop = project_lambert_equal_area
    lim = limit_lambert_equal_area
  else:
    raise ValueError("Unknown projection %s" % projection)

  cart = np.array([pop(v) for v in pts])

  # Convert to polar
  polar = np.array([cart2pol(cp) for cp in cart])

  # Plot
  ax = plt.subplot(111, projection='polar')
  ax.scatter(polar[:,0], polar[:,1], c='k', s=10.0)

  # Make the graph nice
  plt.ylim([0,lim])
  ax.grid(False)
  ax.get_yaxis().set_visible(False)
  if len(axis_labels) == 2:
    plt.xticks([0,np.pi/2], axis_labels)
  else:
    ax.get_xaxis().set_visible(False)
  ax.xaxis.set_minor_locator(plt.NullLocator())

def pol2cart(R, T):
  """
    Convert polar coordinates to cartesian

    Parameters:
      R:         radii
      T:         angle, in radians
  """
  X = R * np.cos(T)
  Y = R * np.sin(T)

  return X, Y

def inverse_pole_figure_discrete(orientations, direction, lattice,  
    reduce_figure = False, color = False,
    sample_symmetry = crystallography.symmetry_rotations("222"),
    x = [1,0,0], y = [0,1,0], axis_labels = None, nline = 100):
  """
    Plot an inverse pole figure given a collection of discrete points.

    Parameters:
      orientations:      list of orientations
      direction:         sample direction
      lattice:           crystal lattice class, including crystal symmetry

    Keyword Args:
      reduce_figure:     reduce to some fundamental region.
                         Options include
                         False -- don't do it
                         "cubic" -- cubic convention
                         [v1,v2,v3] -- list of crystallographic points defining the triangle
      color:             color points based on the provided triangle
      sample_symmetry:   what sample symmetry to apply to the pole figure,
                         defaults to orthorhombic
      x:                 crystallographic x direction for plot, defaults to [1,0,0]
      y:                 crystallographic y direction for plot, defaults to [0,1,0]
      axis_labels:       axis labels to include on the figure
      nline:             number of discrete points to use in plotting lines on the triangle
  """
  pts = np.vstack(tuple(project_ipf(q, lattice, direction, 
    sample_symmetry = sample_symmetry, x = x, y = y) for q in orientations))

  if reduce_figure:
    if reduce_figure == "cubic":
      vs = (np.array([0,0,1.0]), np.array([1.0,0,1]), np.array([1.0,1,1]))
    elif len(reduce_figure) == 3:
      vs = reduce_figure
    else:
      raise ValueError("Unknown reduction type %s!" % reduce_figure)
    
    pts = reduce_points_triangle(pts, v0 = vs[0], v1=vs[1], v2=vs[2])

  pop = project_stereographic
  lim = limit_stereographic

  cpoints = np.array([pop(v) for v in pts])

  # Make the graph nice
  if reduce_figure:
    ax = plt.subplot(111)
    if color:
      rgb = ipf_color(pts, v0 = vs[0], v1 = vs[1], v2=vs[2]) 
      ax.scatter(cpoints[:,0], cpoints[:,1], c=rgb, s = 10.0)
    else:
      ax.scatter(cpoints[:,0], cpoints[:,1], c='k', s = 10.0) 
    ax.axis('off')
    if axis_labels:
      plt.text(0.1,0.11,axis_labels[0], transform = plt.gcf().transFigure)
      plt.text(0.86,0.11,axis_labels[1], transform = plt.gcf().transFigure)
      plt.text(0.74,0.88,axis_labels[2], transform = plt.gcf().transFigure)
    
    for i,j in ((0,1),(1,2),(2,0)):
      v1 = vs[i]
      v2 = vs[j]
      fs = np.linspace(0,1,nline)
      pts = np.array([pop((f*v1+(1-f)*v2)/la.norm(f*v1+(1-f)*v2)) for f in fs])
      plt.plot(pts[:,0], pts[:,1], color = 'k')
  else:
    polar = np.array([cart2pol(v) for v in cpoints])
    ax = plt.subplot(111, projection='polar')
    ax.scatter(polar[:,0], polar[:,1], c='k', s=10.0)
    plt.ylim([0,lim])
    ax.grid(False)
    ax.get_xaxis().set_visible(False)    
    ax.get_yaxis().set_visible(False)
    ax.xaxis.set_minor_locator(plt.NullLocator())

def project_ipf(q, lattice, direction, 
    sample_symmetry = crystallography.symmetry_rotations("222"), 
    x = [1,0,0], y = [0,1,0]):
  """
    Project a single sample direction onto a crystal

    Parameters:
      q:                 lattice orientation
      lattice:           lattice object describing the crystal system
      direction:         sample direction 
    
    Keyword Args:
      sample_symmetry:   sample symmetry operators
      x:                 x direction (crystallographic) of the projection
      y:                 y direction (crystallographic) of the projection
  """
  xv = lattice.miller2cart_direction(x).normalize()
  yv = lattice.miller2cart_direction(y).normalize()

  if not np.isclose(xv.dot(yv),0.0):
    raise ValueError("Lattice directions are not orthogonal!")

  zv = xv.cross(yv)
  
  trans = rotations.Orientation(np.vstack((xv.data,yv.data,zv.data)))
  
  d = tensors.Vector(direction).normalize()

  pts = []
  for srot in sample_symmetry:
    # Sample directions
    spt = srot.apply(d)
    
    # Crystal coordinates
    cpt = q.apply(spt)

    # Lattice symmetry
    for op in lattice.symmetry.ops:
      cppt = op.apply(cpt)

      # Into the right coordinates
      fpt = trans.apply(cppt)

      pts.append(fpt.data)
  
  # By convention just keep the points in the upper hemisphere
  pts = np.array(pts)
  pts = pts[pts[:,2]>0]
  
  return pts

def reduce_points_triangle(pts, v0 = np.array([0,0,1.0]), 
    v1 = np.array([1.0,0,1]), v2 = np.array([1.0,1,1])):
  """
    Reduce points to a standard stereographic triangle

    Default vectors give you the cubic triangle

    Parameters:
      pts:       projected points

    Keyword Args:
      v1:        1st vector
      v2:        2nd vector
      v3:        3rd vector
  """
  v0 /= la.norm(v0)
  v1 /= la.norm(v1)
  v2 /= la.norm(v2)

  n0 = np.cross(v0, v1)
  n1 = np.cross(v1, v2)
  n2 = np.cross(v2, v0)

  npts = []
  for pt in pts:
    if np.dot(pt, n0) >= 0 and np.dot(pt, n1) >= 0 and np.dot(pt, n2) >= 0:
      npts.append(pt)

  return np.array(npts)

def ipf_color(pts, v0 = np.array([0,0,1.0]), 
    v1 = np.array([1.0,0,1]), v2 = np.array([1.0,1,1])):
  """
    Color ipf points based on the provided triangle

    Assumes the points already fall inside the appropriate triangle

    Parameters:
      pts:       list of points

    Keyword Args:
      v0:        point on triangle
      v1:        point on triangle
      v2:        point on triangle
  """
  v0 /= la.norm(v0)
  v1 /= la.norm(v1)
  v2 /= la.norm(v2)

  vs = [v0, v1, v2]

  colors = np.zeros((len(pts),3))

  for i,pt in enumerate(pts):
    for j in range(3):
      colors[i,j] = np.dot(vs[j], pt)

  for i in range(3):
    mf = 1.0
    for j in range(3):
      t = np.dot(vs[i], vs[j])
      if t < mf:
        mf = t
    
    colors[:,i] -= mf
    colors[:,i] /= (1.0 - mf)

  return colors

def ipf_color_chart(v0 = np.array([0,0,1.0]), 
    v1 = np.array([1.0,0,1]), v2 = np.array([1.0,1,1]), 
    axis_labels = None, ngrid = 50, nline = 100):
  """
    Make a color chart for the IPF coloring used here

    Keyword Args:
      v0:            point on triangle
      v1:            point on triangle
      v2:            point on triangle
      axis_labels:   labels to draw on triangle
      ngrid:         number of grid points to use
      nline:         number of line points to use
  """
  vs = [v0,v1,v2]

  pop = project_stereographic
  lim = limit_stereographic
  def ipop(x):
    x1 = inverse_project_stereographic(x).data
    if x1[2] < 0:
      x1[2] *= -1
    return x1

  ax = plt.subplot(111)
  ax.axis('off')
  if axis_labels:
    plt.text(0.12,0.11,axis_labels[0], transform = plt.gcf().transFigure)
    plt.text(0.86,0.11,axis_labels[1], transform = plt.gcf().transFigure)
    plt.text(0.74,0.90,axis_labels[2], transform = plt.gcf().transFigure)

  pts = np.array([pop(v0/la.norm(v0)), pop(v1/la.norm(v1)), pop(v2/la.norm(v2))])
  xrang = np.linspace(np.min(pts[:,0]), np.max(pts[:,0]), ngrid)
  yrang = np.linspace(np.min(pts[:,1]), np.max(pts[:,1]), ngrid)

  X, Y = np.meshgrid(xrang,yrang)
  colors = np.zeros(X.shape+(3,))
  for ind in np.ndindex(X.shape):
    colors[ind] = ipf_color([ipop([X[ind],Y[ind]])], v0 = v0, v1 = v1, v2 = v2)[0]
  
  im = plt.imshow(colors, extent = [xrang[0], xrang[-1], yrang[0], yrang[-1]], 
      origin = 'lower')
  
  net_pts = []
  for i,j in ((0,1),(1,2),(2,0)):
    vi = vs[i]
    vj = vs[j]
    fs = np.linspace(0,1,nline)
    pts = np.array([pop(((1-f)*vi+(f)*vj)/la.norm((1-f)*vi+(f)*vj)) for f in fs])
    net_pts.extend(list(pts[:-1]))

  net_pts = np.array(net_pts)
  pat = Polygon(net_pts, closed = True, edgecolor='k', fc = None, fill = False) 
  ax.add_artist(pat)
  im.set_clip_path(pat)



#!/usr/bin/env python3

from neml import history, interpolate, elasticity
from neml.math import tensors, rotations
from neml.cp import crystallography, slipharden, sliprules, inelasticity, kinematics

import common
from common import differentiate
from nicediff import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonKinematics(object):
  def test_d_stress_rate_d_stress(self):
    nd = diff_symmetric_symmetric(lambda s: self.model.stress_rate(s, 
      self.d, self.w, self.Q, self.H, self.L, self.T, self.fixed), self.S)

    d = self.model.d_stress_rate_d_stress(self.S, self.d, self.w, self.Q, 
        self.H, self.L, self.T, self.fixed)

    self.assertEqual(d, nd)

  def test_d_stress_rate_d_d(self):
    def dfn(d):
      return self.model.stress_rate(self.S, d, self.w, self.Q, self.H, self.L, self.T, 
          self.fixed)

    nd = diff_symmetric_symmetric(dfn, self.d)
    d = self.model.d_stress_rate_d_d(self.S, self.d, self.w, self.Q, self.H, self.L,
        self.T, self.fixed)

    self.assertEqual(nd, d)

  def test_d_stress_rate_d_w(self):
    def dfn(w):
      return self.model.stress_rate(self.S, self.d, w, self.Q, self.H, self.L, self.T, self.fixed)

    nd = diff_symmetric_skew(dfn, self.w)
    d = self.model.d_stress_rate_d_w(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, self.fixed)
   
    self.assertTrue(np.allclose(nd.data, d.data))

  def test_d_stress_rate_d_d_decouple(self):
    def dfn(d):
      fixed = self.model.decouple(self.S, d, self.w, self.Q, self.H, self.L, self.T,
          history.History())
      return self.model.stress_rate(self.S, d, self.w, self.Q, self.H, self.L, self.T, fixed)

    nd = diff_symmetric_symmetric(dfn, self.d) - self.model.d_stress_rate_d_d(self.S, self.d, 
        self.w, self.Q, self.H, self.L, self.T, self.fixed)
    d = self.model.d_stress_rate_d_d_decouple(self.S, self.d, self.w, self.Q, self.H, self.L,
        self.T, self.fixed)

    self.assertTrue(np.allclose(nd.data, d.data, atol = 1.0e-3))

  def test_d_stress_rate_d_w_decouple(self):
    def dfn(w):
      fixed = self.model.decouple(self.S, self.d, w, self.Q, self.H, self.L, self.T,
          history.History())
      return self.model.stress_rate(self.S, self.d, w, self.Q, self.H, self.L, self.T, fixed)

    nd = diff_symmetric_skew(dfn, self.w) - self.model.d_stress_rate_d_w(self.S, self.d,
        self.w, self.Q, self.H, self.L, self.T, self.fixed)
    d = self.model.d_stress_rate_d_w_decouple(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(nd.data, d.data, rtol = 1.0e-3))

  def test_d_stress_rate_d_history(self):
    nd = diff_symmetric_history(lambda h: self.model.stress_rate(self.S, self.d,
      self.w, self.Q, h, self.L, self.T, self.fixed), self.H)
    d = self.model.d_stress_rate_d_history(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(nd, np.array(d).reshape(nd.shape, order = 'F')))

  def test_d_history_rate_d_stress(self):
    nd = diff_history_symmetric(lambda s: self.model.history_rate(s, self.d,
      self.w, self.Q, self.H, self.L, self.T, self.fixed), self.S)
    d = np.array(self.model.d_history_rate_d_stress(self.S, self.d, self.w,
        self.Q, self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape)))

  def test_d_history_rate_d_d(self):
    nd = diff_history_symmetric(lambda d: self.model.history_rate(self.S, d,
      self.w, self.Q, self.H, self.L, self.T, self.fixed), self.d)
    d = np.array(self.model.d_history_rate_d_d(self.S, self.d, self.w, self.Q,
        self.H, self.L, self.T,self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape)))

  def test_d_history_rate_d_w(self):
    nd = diff_history_skew(lambda w: self.model.history_rate(self.S, self.d,
      w, self.Q, self.H, self.L, self.T, self.fixed), self.w)
    d = np.array(self.model.d_history_rate_d_w(self.S, self.d, self.w, self.Q,
      self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape)))

  def test_d_history_rate_d_d_decouple(self):
    def dfn(d):
      fixed = self.model.decouple(self.S, d, self.w, self.Q, self.H, self.L, self.T,
          history.History())
      return self.model.history_rate(self.S, d, self.w, self.Q, self.H, self.L, self.T, fixed)
    
    nd1 = diff_history_symmetric(dfn, self.d)

    nd = nd1 - np.array(self.model.d_history_rate_d_d(self.S, self.d, self.w,
        self.Q, self.H, self.L, self.T, self.fixed)).reshape(nd1.shape, order = 'F')
    d = np.array(self.model.d_history_rate_d_d_decouple(self.S, self.d, self.w, self.Q,
        self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape, order = 'F')))
  
  def test_d_history_rate_d_w_decouple(self):
    def dfn(w):
      fixed = self.model.decouple(self.S, self.d, w, self.Q, self.H, self.L, self.T,
          history.History())
      return self.model.history_rate(self.S, self.d, w, self.Q, self.H, self.L, self.T, fixed)
  
    nd1 = diff_history_skew(dfn, self.w)

    nd = nd1 - np.array(self.model.d_history_rate_d_w(self.S, self.d, self.w,
        self.Q, self.H, self.L, self.T, self.fixed)).reshape(nd1.shape, order = 'F')
    d = np.array(self.model.d_history_rate_d_w_decouple(self.S, self.d, self.w, self.Q,
        self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape, order = 'F')))

  def test_d_history_rate_d_history(self):
    nd = diff_history_history(lambda h: self.model.history_rate(self.S, self.d,
      self.w, self.Q, h, self.L, self.T, self.fixed), self.H)
    d = np.array(self.model.d_history_rate_d_history(self.S, self.d, self.w,
      self.Q, self.H, self.L, self.T, self.fixed))

    self.assertTrue(np.allclose(nd, d.reshape(nd.shape)))

class TestStandardKinematics(unittest.TestCase, CommonKinematics):
  def setUp(self):
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.dn = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.dn = 0.5*(self.dn + self.dn.T)
    self.d = tensors.Symmetric(self.dn)

    self.wn = np.array([[-9.36416517,  2.95527444,  8.70983194],
           [-1.54693052,  8.7905658 , -5.10895168],
           [-8.52740468, -0.7741642 ,  2.89544992]])
    self.wn = 0.5 * (self.wn - self.wn.T)
    self.w = tensors.Skew(self.wn)

    self.model = kinematics.StandardKinematicModel(self.emodel, self.imodel)
    
    self.fspin = self.model.spin(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, history.History())

    self.fixed = self.model.decouple(self.S, self.d, self.w, self.Q, self.H, self.L, self.T,
        history.History())

  def test_setup_history(self):
    H1 = history.History()
    self.model.populate_history(H1)
    self.model.init_history(H1)

    H2 = history.History()
    self.imodel.populate_history(H2)
    self.imodel.init_history(H2)

    self.assertTrue(np.allclose(np.array(H1), np.array(H2)))

  def test_stress_rate(self):
    Cfull = ms2ts(self.emodel.C_tensor(self.T, self.Q).data.reshape((6,6)))
    Sfull = ms2ts(self.emodel.S_tensor(self.T, self.Q).data.reshape((6,6)))
    d = usym(self.d.data)
    w = uskew(self.w.data)
    Ofull = uskew(self.fspin.data)
    
    dp = usym(self.imodel.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed).data)
    wp = uskew(self.imodel.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed).data)

    O = wp + Ofull

    stress = usym(self.S.data)
    
    e = np.einsum('ijkl,kl', Sfull, stress)

    sdot1 = tensors.Symmetric(
        np.einsum('ijkl,kl', Cfull, d - dp - np.dot(e, O) + np.dot(O, e)))
    
    sdot2 = self.model.stress_rate(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, self.fixed)

    self.assertEqual(sdot1, sdot2)

  def test_hist_rate(self):
    H1 = self.model.history_rate(self.S, self.d, self.w,
        self.Q, self.H, self.L, self.T, self.fixed)
    H2 = self.imodel.history_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)

    self.assertTrue(np.allclose(np.array(H1), np.array(H2)))

  def test_spin_rate(self):
    Cfull = ms2ts(self.emodel.C_tensor(self.T, self.Q).data.reshape((6,6)))
    Sfull = ms2ts(self.emodel.S_tensor(self.T, self.Q).data.reshape((6,6)))
    d = usym(self.d.data)
    w = uskew(self.w.data)
    Ofull = uskew(self.fspin.data)
    
    dp = usym(self.imodel.d_p(self.S, self.Q, self.H, self.L, self.T, self.fixed).data)
    wp = uskew(self.imodel.w_p(self.S, self.Q, self.H, self.L, self.T, self.fixed).data)

    O = wp + Ofull

    stress = usym(self.S.data)
    
    e = np.einsum('ijkl,kl', Sfull, stress)

    spin1 = tensors.Skew(
        w - wp - np.dot(e, dp) + np.dot(dp, e))
    spin2 = self.model.spin(self.S, self.d, self.w, self.Q, self.H, 
        self.L, self.T, self.fixed)

    self.assertTrue(spin1, spin2)

class TestStandardKinematicsComplicated(unittest.TestCase, CommonKinematics):
  def setUp(self):
    self.strength_0 = 35.0
    self.H = history.History()
    self.H.add_scalar("strength0")
    self.H.set_scalar("strength0", self.strength_0)

    self.strength_1 = 25.0
    self.H.add_scalar("strength1")
    self.H.set_scalar("strength1", self.strength_1)

    self.tau0_0 = 10.0
    self.tau_sat_0 = 50.0
    self.b_0 = 2.5

    self.tau0_1 = 5.0
    self.tau_sat_1 = 25.0
    self.b_1 = 1.0

    self.strengthmodel = slipharden.SumSlipSingleStrengthHardening(
        [
          slipharden.VoceSlipHardening(self.tau_sat_0, self.b_0, self.tau0_0),
          slipharden.VoceSlipHardening(self.tau_sat_1, self.b_1, self.tau0_1)
          ])
    
    self.g0 = 1.0
    self.n = 3.0
    self.slipmodel = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.imodel = inelasticity.AsaroInelasticity(self.slipmodel)

    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))

    self.T = 300.0

    self.mu = 29000.0
    self.E = 120000.0
    self.nu = 0.3

    self.emodel = elasticity.CubicLinearElasticModel(self.E, 
        self.nu, self.mu, "moduli")

    self.dn = np.array([[4.1,2.8,-1.2],[3.1,7.1,0.2],[4,2,3]])
    self.dn = 0.5*(self.dn + self.dn.T)
    self.d = tensors.Symmetric(self.dn)

    self.wn = np.array([[-9.36416517,  2.95527444,  8.70983194],
           [-1.54693052,  8.7905658 , -5.10895168],
           [-8.52740468, -0.7741642 ,  2.89544992]])
    self.wn = 0.5 * (self.wn - self.wn.T)
    self.w = tensors.Skew(self.wn)

    self.model = kinematics.StandardKinematicModel(self.emodel, self.imodel)
    
    self.fspin = self.model.spin(self.S, self.d, self.w, self.Q, self.H,
        self.L, self.T, history.History())

    self.fixed = self.model.decouple(self.S, self.d, self.w, self.Q, self.H, self.L, self.T,
        history.History())



import sys
sys.path.append("..")

from neml import models, elasticity, ri_flow, hardening, surfaces, visco_flow, general_flow, creep, damage
from common import *

import unittest

class TestSimpleMaterials(unittest.TestCase):
  def setUp(self):
    self.G1 = 36000.0
    self.G2 = 25000.0

    self.K1 = 10000.0
    self.K2 = 12000.0

    self.elastic1 = elasticity.IsotropicLinearElasticModel(
        self.G1, "shear",
        self.K1, "bulk")
    self.elastic2 = elasticity.IsotropicLinearElasticModel(
        self.G2, "shear",
        self.K2, "bulk")
    
  def are_equal(self, e1, e2, T = 30):
    self.assertTrue(np.allclose(e1.C(T), e2.C(T)))

  def test_linear_elastic(self):
    model = models.SmallStrainElasticity(self.elastic1)
    self.are_equal(self.elastic1, model.elastic)
    model.set_elastic_model(self.elastic2)
    self.are_equal(self.elastic2, model.elastic)

  def test_perfect_plasicity(self):
    surface = surfaces.IsoJ2()
    sy = 100.0
    model = models.SmallStrainPerfectPlasticity(self.elastic1, surface, sy)
    self.are_equal(self.elastic1, model.elastic)
    model.set_elastic_model(self.elastic2)
    self.are_equal(self.elastic2, model.elastic)

  def test_ri_plasticity(self):
    surface = surfaces.IsoJ2()
    sy = 100.0
    K = 1000.0
    H = 500.0
    hrule = hardening.CombinedHardeningRule(
        hardening.LinearIsotropicHardeningRule(sy, K),
        hardening.LinearKinematicHardeningRule(H))
    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)
    model = models.SmallStrainRateIndependentPlasticity(self.elastic1, 
        flow)
    self.are_equal(self.elastic1, model.elastic)
    model.set_elastic_model(self.elastic2)
    self.are_equal(self.elastic2, model.elastic)

class TestCompoundMaterials(unittest.TestCase):
  def setUp(self):
    self.G1 = 36000.0
    self.G2 = 25000.0

    self.K1 = 10000.0
    self.K2 = 12000.0

    self.elastic1 = elasticity.IsotropicLinearElasticModel(
        self.G1, "shear",
        self.K1, "bulk")
    self.elastic2 = elasticity.IsotropicLinearElasticModel(
        self.G2, "shear",
        self.K2, "bulk")

    self.emodel1 = models.SmallStrainElasticity(self.elastic1)
    self.emodel2 = models.SmallStrainElasticity(self.elastic2)

    self.tiny_strain = np.array([1e-6,0,0,0,0,0])
    self.T = 300.0
    self.t = 1e-9

  def very_close(self, model1, model2):
    self.assertTrue(np.allclose(self.small_step(model1), 
      self.small_step(model2)))

  def small_step(self, model):
    h0 = model.init_store()
    s0 = np.zeros((6,))
    e0 = np.zeros((6,))
    t0 = 0.0
    u0 = 0.0
    p0 = 0.0

    stress, hist, A, u, p = model.update_sd(
        self.tiny_strain, e0, self.T, self.T, self.t, t0,
        s0, h0, u0, p0)

    return stress
  
  def test_comparison(self):
    self.very_close(self.emodel1, self.emodel1)

  def test_creep_plasticity(self):
    surface = surfaces.IsoJ2()
    sy = 100.0
    bmodel = models.SmallStrainPerfectPlasticity(self.elastic1, surface, sy)
    A = 1.85e-10
    n = 12.0
    
    smodel = creep.PowerLawCreep(A, n)
    cmodel = creep.J2CreepModel(smodel)
    model = models.SmallStrainCreepPlasticity(self.elastic1, bmodel, cmodel)

    self.very_close(model, self.emodel1)

    model.set_elastic_model(self.elastic2)

    self.very_close(model, self.emodel2)

  def test_rd(self):
    s0 = 10.0
    R = 150.0
    d = 10.0

    n = 3.0
    eta = 200.0

    surface = surfaces.IsoJ2()
    hrule = hardening.VoceIsotropicHardeningRule(s0, R, d)
    g = visco_flow.GPowerLaw(n, eta)

    vmodel = visco_flow.PerzynaFlowRule(surface, hrule, g)
    flow = general_flow.TVPFlowRule(self.elastic1, vmodel)
    model = models.GeneralIntegrator(self.elastic1, flow)

    self.very_close(model, self.emodel1)
    
    model.set_elastic_model(self.elastic2)

    self.very_close(model, self.emodel2)

  def test_damage(self):
    s0 = 180.0
    Kp = 1000.0
    H = 1000.0

    surface = surfaces.IsoKinJ2()
    iso = hardening.LinearIsotropicHardeningRule(s0, Kp)
    kin = hardening.LinearKinematicHardeningRule(H)
    hrule = hardening.CombinedHardeningRule(iso, kin)

    flow = ri_flow.RateIndependentAssociativeFlow(surface, hrule)

    bmodel = models.SmallStrainRateIndependentPlasticity(self.elastic1, 
        flow)

    W0 = 10.0
    k0 = 0.0001
    a0 = 2.0

    model1 = damage.NEMLExponentialWorkDamagedModel_sd(
        self.elastic1, W0, k0, 
        a0, bmodel)

    W02 = 10.0
    k02 = 0.001
    a02 = 1.5

    model2 = damage.NEMLExponentialWorkDamagedModel_sd(
        self.elastic1, W02, k02, 
        a02, bmodel)

    model = damage.CombinedDamageModel_sd(self.elastic1, 
        [model1, model2], bmodel)
    
    self.very_close(model, self.emodel1)
    
    model.set_elastic_model(self.elastic2)

    self.very_close(model, self.emodel2)



import sys
sys.path.append('..')

from neml import solvers, creep, elasticity, interpolate
import unittest

from common import *

import numpy as np
import numpy.linalg as la

class CommonScalarCreep(object):
  """
    Common tests to impose on scalar creep laws.
  """
  def test_dg_ds(self):
    """
      Derivative of creep rate wrt. effective stress
    """
    dfn = lambda x: self.model.g(x, self.e, self.t, self.T)
    nderiv = differentiate(dfn, self.s)
    cderiv = self.model.dg_ds(self.s, self.e, self.t, self.T)
    self.assertTrue(np.isclose(nderiv, cderiv, rtol = 1.0e-4))

  def test_dg_de(self):
    """
      Derivative of creep rate wrt. effective strain
    """
    dfn = lambda x: self.model.g(self.s, x, self.t, self.T)
    nderiv = differentiate(dfn, self.e)
    cderiv = self.model.dg_de(self.s, self.e, self.t, self.T)
    self.assertTrue(np.isclose(nderiv, cderiv, rtol = 1.0e-4))

  def test_dg_dt(self):
    """
      Derivative of creep rate wrt. time
    """
    dfn = lambda x: self.model.g(self.s, self.e, x, self.T)
    nderiv = differentiate(dfn, self.t)
    cderiv = self.model.dg_dt(self.s, self.e, self.t, self.T)
    self.assertTrue(np.isclose(nderiv, cderiv))

  def test_dg_dT(self):
    """
      Derivative of creep rate wrt. temperature
    """
    dfn = lambda x: self.model.g(self.s, self.e, self.t, x)
    nderiv = differentiate(dfn, self.T)
    cderiv = self.model.dg_dT(self.s, self.e, self.t, self.T)
    if np.isclose(cderiv, 0.0):
      self.assertTrue(True) # This just means I don't care
    else:
      self.assertTrue(np.isclose(nderiv, cderiv))

class TestPowerLawCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.A = 1.0e-6
    self.n = 5.0

    self.model = creep.PowerLawCreep(self.A, self.n) 

    self.T = 300.0
    self.e = 0.1
    self.s = 150.0
    self.t = 10.0

  def test_properties(self):
    self.assertTrue(np.isclose(self.A, self.model.A(self.T)))
    self.assertTrue(np.isclose(self.n, self.model.n(self.T)))

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    g_calc = self.A * self.s ** (self.n)
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestNormalizedPowerLawCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.s0 = 500.0
    self.n = 3.0

    self.model = creep.NormalizedPowerLawCreep(self.s0, self.n) 

    self.T = 300.0
    self.e = 0.1
    self.s = 150.0
    self.t = 10.0

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    g_calc = (self.s / self.s0) ** (self.n)
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestBlackburnMinimumCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.A = 1.0e-6
    self.beta = -2.0e-2
    self.n = 5.0
    self.Q = 67000.0
    self.R = 1.987

    self.model = creep.BlackburnMinimumCreep(self.A, self.n, self.beta,
        self.R, self.Q) 

    self.T = 300.0
    self.e = 0.1
    self.s = 150.0
    self.t = 10.0

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    g_calc = self.A * np.sinh(self.beta*self.s/self.n)**self.n * np.exp(-self.Q/(self.R*self.T))
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestSwindemanMinimumCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.C = 2.25e20
    self.n = 5.0
    self.V = 0.038
    self.Q = 77280.0

    self.model = creep.SwindemanMinimumCreep(self.C, self.n, self.V,
        self.Q) 

    self.T = 600+273.15
    self.e = 0.1
    self.s = 150.0
    self.t = 10.0

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    g_calc = self.C * self.s ** self.n * np.exp(self.V * self.s) * np.exp(-self.Q/self.T)
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestRegionKMCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.b = 2.019 * 1.0e-7
    self.eps0 = 1.0e6
    self.E = 160000.0
    self.nu = 0.3

    self.emodel = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")

    self.kboltz = 1.38064e-23 * 1000.0
    
    self.cuts = [0.00145709]
    self.A = [-0.7869, -0.1695]
    self.B = [-4.2388, -0.2065]

    self.model = creep.RegionKMCreep(self.cuts, self.A, self.B, self.kboltz,
        self.b, self.eps0, self.emodel)

    self.T = 300.0
    self.e = 0.1
    self.s = 300.0
    self.t = 10.0

class TestNortonBaileyCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.A = 1.0e-6
    self.m = 0.25
    self.n = 5.0

    self.model = creep.NortonBaileyCreep(self.A, self.m, self.n) 

    self.T = 300.0
    self.e = 0.1
    self.s = 150.0
    self.t = 10.0

  def test_properties(self):
    self.assertTrue(np.isclose(self.A, self.model.A(self.T)))
    self.assertTrue(np.isclose(self.m, self.model.m(self.T)))
    self.assertTrue(np.isclose(self.n, self.model.n(self.T)))

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    g_calc = self.m * self.A**(1.0/self.m) * self.s**(self.n/self.m) * self.e**((self.m-1.0)/self.m)
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestMukherjeeCreep(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.A = 1.0e10     # Unitless
    self.n = 7.9        # Unitless
    self.D0 = 3.7e-5    # m^2 / s
    self.Qv = 280.0     # kJ/ mol
    self.b = 2.28e-10   # m

    self.E = 100000.0
    self.nu = 0.3
    self.emodel = elasticity.IsotropicLinearElasticModel(self.E, "youngs",
        self.nu, "poissons")
    
    self.k = 1.380e-23 * 1.0e-6 # m^3 * MPa / K
    self.R = 8.314 / 1000 # kJ / (K * mol)

    self.t = 10.0
    self.e = 0.1
    self.s = 150.0
    self.T = 550.0 + 273.15

    self.model = creep.MukherjeeCreep(self.emodel, self.A, self.n, self.D0,
        self.Qv, self.b, self.k, self.R)

  def test_properties(self):
    self.assertTrue(np.isclose(self.A, self.model.A))
    self.assertTrue(np.isclose(self.n, self.model.n))
    self.assertTrue(np.isclose(self.D0, self.model.D0))
    self.assertTrue(np.isclose(self.Qv, self.model.Q))
    self.assertTrue(np.isclose(self.b, self.model.b))
    self.assertTrue(np.isclose(self.k, self.model.k))
    self.assertTrue(np.isclose(self.R, self.model.R))

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    mu = self.emodel.G(self.T)
    g_calc = self.A * self.D0 * np.exp(-self.Qv / (self.R * self.T)
        ) * mu * self.b / (self.k * self.T) * (self.s / mu) ** self.n
    self.assertTrue(np.isclose(g_direct, g_calc))

class TestGenericCreep(unittest.TestCase, CommonScalarCreep):
  """
    Tests GenericCreep --- creep law given as a generic interpolation
    of log stress.
  """
  def setUp(self):
    # Polynomial coefficients: log(creep rate) = np.polyval(poly, log(stress))
    self.poly = [2.5, -50.0, 365.0, -1200.0, 1500.0]
    self.ifn = interpolate.PolynomialInterpolate(self.poly)
    
    # Model only has one input parameter (the interpolation formula)
    self.model = creep.GenericCreep(self.ifn)
    
    self.T = 300.0  # Temperature to test at
    self.e = 0.1    # Effective strain to test at
    self.s = 150.0  # Effective stress to test at
    self.t = 10.0   # Time to test at

  def test_g(self):
    """
      Compare the creep rate calculated "by hand" to the creep rate
      returned by the model.
    """
    g_calc = np.exp(np.polyval(self.poly, np.log(self.s)))
    g_model = self.model.g(self.s, self.e, self.t, self.T)
    self.assertTrue(np.isclose(g_calc, g_model))

class Test225CreepCase1(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.model = creep.MinCreep225Cr1MoCreep() 

    self.T = 500 + 273.15
    self.e = 0.1
    self.s = 50.0
    self.t = 10.0

    self.U = interpolate.PiecewiseLinearInterpolate(
        [371, 400, 450, 500, 550, 600, 621, 649],
        [471,468,452,418,634,284,300,270])

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    Uv = self.U.value(self.T - 273.15)
    g_calc = 10.0**(6.7475+0.011426*self.s+987.72/Uv * np.log10(self.s) - 13494.0 / self.T) / 100.0
    self.assertTrue(np.isclose(g_direct, g_calc))

class Test225CreepCase2(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.model = creep.MinCreep225Cr1MoCreep() 

    self.T = 500 + 273.15
    self.e = 0.1
    self.s = 100.0
    self.t = 10.0

    self.U = interpolate.PiecewiseLinearInterpolate(
        [371, 400, 450, 500, 550, 600, 621, 649],
        [471,468,452,418,634,284,300,270])

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    Uv = self.U.value(self.T - 273.15)
    g_calc = 10.0**(6.7475+0.011426*self.s+987.72/Uv * np.log10(self.s) - 13494.0 / self.T) / 100.0
    self.assertTrue(np.isclose(g_direct, g_calc))

class Test225CreepCase3(unittest.TestCase, CommonScalarCreep):
  def setUp(self):
    self.model = creep.MinCreep225Cr1MoCreep() 

    self.T = 600 + 273.15
    self.e = 0.1
    self.s = 100.0
    self.t = 10.0

    self.U = interpolate.PiecewiseLinearInterpolate(
        [371, 400, 450, 500, 550, 600, 621, 649],
        [471,468,452,418,634,284,300,270])

  def test_g(self):
    g_direct = self.model.g(self.s, self.e, self.t, self.T)
    Uv = self.U.value(self.T - 273.15)
    g_calc = 10.0**(11.498 - 8.2226 * Uv / self.T - 20448 / self.T + 
        5862.4 / self.T * np.log10(self.s)) / 100.0
    self.assertTrue(np.isclose(g_direct, g_calc))

class CommonCreepModel(object):
  """
    Tests common to all creep models
  """
  def test_tangent(self):
    strain = np.copy(self.e)
    stress = np.copy(self.s)

    ts = np.linspace(0,self.t)

    for i in range(1, len(ts)):
      t_np1 = ts[i]
      t_n = ts[i-1]
      s_np1 = stress
      e_next, A_next = self.model.update(stress, strain, self.T, self.T,
          t_np1, t_n)

      dfn = lambda x: self.model.update(x, strain, self.T, self.T, t_np1,
          t_n)[0]
      nA = differentiate(dfn, s_np1)
      self.assertTrue(np.allclose(nA, A_next, rtol = 1.0e-4))

      strain = np.copy(e_next)

  def test_jacobian(self):
    ts = self.model.make_trial_state(self.s, self.e, self.T, self.T,
      self.dt + self.t, self.t)
    R, J = self.model.RJ(self.x, ts)
    nJ = differentiate(lambda x: self.model.RJ(x, ts)[0], self.x)
    self.assertTrue(np.allclose(J, nJ, rtol = 1.0e-4))

  def test_df_ds(self):
    dfn = lambda x: self.model.f(x, self.e, self.t, self.T)
    nderiv = differentiate(dfn, self.s)
    cderiv = self.model.df_ds(self.s, self.e, self.t, self.T)
    self.assertTrue(np.allclose(nderiv, cderiv, rtol = 1.0e-4))

  def test_df_de(self):
    dfn = lambda x: self.model.f(self.s, x, self.t, self.T)
    nderiv = differentiate(dfn, self.e)
    cderiv = self.model.df_de(self.s, self.e, self.t, self.T)
    self.assertTrue(np.allclose(nderiv, cderiv, rtol = 1.0e-4))

  def test_df_dt(self):
    dfn = lambda x: self.model.f(self.s, self.e, x, self.T)
    nderiv = differentiate(dfn, self.t)
    cderiv = self.model.df_dt(self.s, self.e, self.t, self.T)
    self.assertTrue(np.allclose(nderiv, cderiv))

  def test_df_dT(self):
    dfn = lambda x: self.model.f(self.s, self.e, self.t, x)
    nderiv = differentiate(dfn, self.T)
    cderiv = self.model.df_dT(self.s, self.e, self.t, self.T)
    self.assertTrue(np.allclose(nderiv, cderiv))

class TestJ2Creep(unittest.TestCase, CommonCreepModel):
  def setUp(self):
    self.A = 1.0e-10
    self.m = 0.25
    self.n = 5.0

    self.smodel = creep.NortonBaileyCreep(self.A, self.m, self.n)

    self.model = creep.J2CreepModel(self.smodel)

    self.s = np.array([100.0,-25.0,-5.0, 20.0,15.0,3.0])
    self.e = np.array([0.05, -0.01, -0.01, 0.025, 0.03, -0.01])
    self.x = np.array([0.06, 0.01, 0.02, 0.01, 0.02, -0.05])
    self.T = 300.0
    self.t = 10.0
    self.dt = 2.0

    tr = np.sum(self.s[:3])
    self.sdev = self.s - np.array([1,1,1,0,0,0]) * tr / 3.0
    self.se = np.sqrt(3.0/2.0) * la.norm(self.sdev)
    self.ee = np.sqrt(2.0/3.0) * la.norm(self.e)

  def test_f(self):
    f_direct = self.model.f(self.s, self.e, self.t, self.T)
    g_direct = self.smodel.g(self.se, self.ee, self.t, self.T)
    f_calc = g_direct * 3.0 / 2.0 * self.sdev / self.se

    self.assertTrue(np.allclose(f_direct, f_calc))

  def test_f_uni(self):
    """
      Make sure all our "effectives" work out correctly
    """
    s = np.array([100.0, 0, 0, 0, 0, 0])
    e = np.array([0.1, -0.05, -0.05, 0, 0, 0])
    f_direct = self.model.f(s, e, self.t, self.T)
    
    sdev = s - np.array([1,1,1,0,0,0]) * np.sum(s[:3]) / 3.0
    se = np.sqrt(3.0/2.0) * la.norm(sdev)
    ee = np.sqrt(2.0/3.0) * la.norm(e)

    g_direct = self.smodel.g(se, ee, self.t, self.T)
    
    self.assertTrue(np.isclose(g_direct, f_direct[0]))

    self.assertTrue(np.isclose(-g_direct/2.0, f_direct[1]))
    self.assertTrue(np.isclose(-g_direct/2.0, f_direct[2]))

    self.assertTrue(np.allclose([0,0,0], f_direct[3:]))





# This file lightly adapted from the example CMake build provided in the pybind11 distribution
# located at https://github.com/pybind/cmake_example/blob/master/setup.py
# That file is released under an open source license, contained in the pybind11 subdirectory
# of this package.

import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding = 'utf-8') as f:
  long_description = f.read()

class CMakeExtension(Extension):
  def __init__(self, name, sourcedir=''):
      Extension.__init__(self, name, sources=[])
      self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
  def run(self):
    try:
      out = subprocess.check_output(['cmake', '--version'])
    except OSError:
      raise RuntimeError("CMake must be installed to build the following extensions: " +
                          ", ".join(e.name for e in self.extensions))

    if platform.system() == "Windows":
      cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
      if cmake_version < '3.1.0':
        raise RuntimeError("CMake >= 3.1.0 is required on Windows")

    for ext in self.extensions:
        self.build_extension(ext)

  def build_extension(self, ext):
    extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
    # required for auto-detection of auxiliary "native" libs
    if not extdir.endswith(os.path.sep):
      extdir += os.path.sep

    cmake_args = [
                  '-DWRAP_PYTHON=ON',
                  '-DUSE_OPENMP=OFF',
                  '-DBUILD_UTILS=OFF',
                  '-DINSTALL=' + os.path.join(extdir,"neml")]

    cfg = 'Debug' if self.debug else 'Release'
    build_args = ['--config', cfg]

    if platform.system() == "Windows":
      cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
      if sys.maxsize > 2**32:
          cmake_args += ['-A', 'x64']
      build_args += ['--', '/m']
    else:
      cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
      build_args += ['--', '-j2']

    env = os.environ.copy()
    env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                          self.distribution.get_version())
    if not os.path.exists(self.build_temp):
        os.makedirs(self.build_temp)
  
    subprocess.check_call(['cmake', '.'] + cmake_args, env=env)
    subprocess.check_call(['cmake', '--build', '.'] + build_args)

setup (
    # Name of the project
    name = 'neml',
    # Version
    version = '1.3.2',
    # One line-description
    description = "Nuclear Engineering Material model Library",
    # README
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    # Project webpage
    url='https://github.com/Argonne-National-Laboratory/neml',
    # Author
    author='Argonne National Laboratory',
    # email
    author_email = 'messner@anl.gov',
    # Various things for pypi
    classifiers=[
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: C++',
      'Programming Language :: Python :: 3',
      'Operating System :: OS Independent'
      ],
    # Which version of python is needed
    python_requires='>=3.6',
    # Keywords to help find
    keywords='materials structures modeling',
    # Definitely not zip safe
    zip_safe=False,
    # Require the various modules
    ext_modules=[CMakeExtension('neml')],
    # Add the CMake builder
    cmdclass=dict(build_ext=CMakeBuild),
    # Get the python files
    packages=find_packages(),
    # Locate tests
    test_suite='nose.collector',
    tests_required=['nose'],
    # Python dependencies
    install_requires=[
      'numpy',
      'scipy',
      'matplotlib',
      'networkx'
      ],
)



from neml import models, parse, drivers

import os.path

import unittest, pickle
import numpy as np

from common import *

xml = localize("regression/reference.xml")
models = localize("regression/models.txt")

strain_rate = 1.0e-4

def test_all_regression():
  names = [line.strip() for line in open(models)]
  for model in names:
    # Too expensive
    if model == "linearcp":
      continue
    yield check_regression, model

def check_regression(model):
  nmodel = parse.parse_xml(xml, model)
  res = drivers.uniaxial_test(nmodel, strain_rate)
  reference = pickle.load(open(os.path.join('test/regression',
    model+'.pickle'), 'rb'))
  assert(np.allclose(res['strain'], reference['strain']))
  assert(np.allclose(res['stress'], reference['stress']))
  assert(np.allclose(res['energy_density'], reference['energy_density']))
  assert(np.allclose(res['plastic_work'], reference['plastic_work']))




import sys
sys.path.append('..')

from neml import models, elasticity, surfaces, hardening, ri_flow
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonLD(object):
  """
    Generic tests that can be applied to arbitrary models
  """
  def test_tangent_at_points(self):
    for c in self.conditions:
      hist_n = c['hist_n']
      stress_n = c['stress_n']
      d_n = c['d_n']
      d_np1 = c['d_np1']
      w_n = c['w_n']
      w_np1 = c['w_np1']
      t_n = 0.0
      t_np1 = c['dt']
      T_n = c['T']
      T_np1 = c['T']
      u_n = 0.0
      p_n = 0.0

      stress_np1, hist_np1, A_np1, B_np1, u_np1, p_np1 = self.model.update_ld_inc(
          d_np1, d_n, w_np1, w_n, T_np1, T_n, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)

      ufn_d = lambda d: self.model.update_ld_inc(
          d, d_n, w_np1, w_n, T_np1, T_n, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)[0]
      ufn_w = lambda w: self.model.update_ld_inc(
          d_np1, d_n, w, w_n, T_np1, T_n, t_np1, t_n, stress_n, hist_n,
          u_n, p_n)[0]

      A_num = differentiate(ufn_d, d_np1)
      B_num = differentiate(ufn_w, w_np1)
      
      print("A")
      print((A_np1-A_num)/A_num)

      print("B")
      print((B_np1-B_num)/B_num)

      self.assertTrue(np.allclose(A_np1, A_num, rtol = 1.0e-2))
      self.assertTrue(np.allclose(B_np1, B_num, rtol = 1.0e-2))

  def test_tangent_proportional(self):
    for dirs in self.directions:
      ddir = dirs['d']
      wdir = dirs['w']

      t_n = 0.0
      hist_n = self.model.init_store()
      d_n = np.zeros((6,))
      w_n = np.zeros((3,))
      stress_n = np.zeros((6,))

      for i in range(self.nsteps):
        t_np1 = t_n + self.dt
        d_np1 = d_n + ddir
        w_np1 = w_n + wdir

        stress_np1, hist_np1, A_np1, B_np1, u_np1, p_np1 = self.model.update_ld_inc(
            d_np1, d_n, w_np1, w_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
            0.0, 0.0)

        ufn_d = lambda d: self.model.update_ld_inc(
            d, d_n, w_np1, w_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
            0.0, 0.0)[0]
        ufn_w = lambda w: self.model.update_ld_inc(
            d_np1, d_n, w, w_n, self.T, self.T, t_np1, t_n, stress_n, hist_n,
            0.0, 0.0)[0]

        A_num = differentiate(ufn_d, d_np1)
        B_num = differentiate(ufn_w, w_np1)

        self.assertTrue(np.allclose(A_num, A_np1, rtol = 1.0e-2))
        self.assertTrue(np.allclose(B_num, B_np1, rtol = 1.0e-2))

        t_n = t_np1
        d_n = np.copy(d_np1)
        w_n = np.copy(w_np1)
        stress_n = np.copy(stress_np1)
        hist_n = np.copy(hist_np1)


class TestLinearElastic(CommonLD, unittest.TestCase):
  def setUp(self):
    self.E = 100000.0
    self.nu = 0.29

    self.elastic = elasticity.IsotropicLinearElasticModel(self.E,
        "youngs", self.nu, "poissons")
    self.model = models.SmallStrainElasticity(self.elastic)
    
    self.conditions = [
        {'hist_n': np.zeros((6,)),
          'stress_n': np.zeros((6,)),
          'd_n': np.zeros((6,)),
          'd_np1': np.array([0.1,0.05,-0.025,0.15,0.2,-0.05]),
          'w_n': np.zeros((3,)),
          'w_np1': np.array([-0.15,0.1,0.05]),
          'dt': 1.0,
          'T': 300.0},
        {'hist_n': np.zeros((6,)),
          'stress_n': np.array([10.0,-5.0,30.0,-5.0,10.0,15.0]),
          'd_n': np.zeros((6,)),
          'd_np1': np.array([0.05,0.5,0.25,0.20,-0.2,0.25]),
          'w_n': np.zeros((3,)),
          'w_np1': np.array([-0.15,0.25,0.05]),
          'dt': 1.0,
          'T': 300.0}
        ]

    self.directions = [
        {'d': np.array([0.1,-0.15,0.2,-0.05,0.15,0.25]),
          'w': np.array([0.25,-0.15,0.1])}]

    self.nsteps = 10
    self.dt = 1.0
    self.T = 300.0

class TestSimplePlastic(CommonLD, unittest.TestCase):
  def setUp(self):
    self.E = 100000.0
    self.nu = 0.29
    self.sY = 100.0
    self.H = 1200.0

    elastic = elasticity.IsotropicLinearElasticModel(self.E,
        "youngs", self.nu, "poissons")
    surf = surfaces.IsoJ2()
    hard = hardening.LinearIsotropicHardeningRule(self.sY, self.H)
    flow = ri_flow.RateIndependentAssociativeFlow(surf, hard)
    self.model = models.SmallStrainRateIndependentPlasticity(elastic, flow)

    self.conditions = [
        {'hist_n': self.model.init_store(),
          'stress_n': np.zeros((6,)),
          'd_n': np.zeros((6,)),
          'd_np1': np.array([0.1,0.05,-0.025,0.15,0.2,-0.05]),
          'w_n': np.zeros((3,)),
          'w_np1': np.array([-0.15,0.1,0.05]),
          'dt': 1.0,
          'T': 300.0},
        {'hist_n': self.model.init_store(),
          'stress_n': np.array([10.0,-5.0,30.0,-5.0,10.0,15.0]),
          'd_n': np.zeros((6,)),
          'd_np1': np.array([0.05,0.5,0.25,0.20,-0.2,0.25]),
          'w_n': np.zeros((3,)),
          'w_np1': np.array([-0.15,0.25,0.05]),
          'dt': 1.0,
          'T': 300.0}
        ]

    self.directions = [
        {'d': np.array([0.1,-0.15,0.2,-0.05,0.15,0.25]),
          'w': np.array([0.25,-0.15,0.1])}]

    self.nsteps = 10
    self.dt = 1.0
    self.T = 300.0



import sys
sys.path.append('..')

from neml import hardening, interpolate
import unittest

from common import *

import numpy as np
import numpy.linalg as la

class CommonHardening(object):
  """
    Tests that can apply to all hardening rules
  """
  def test_history(self):
    self.assertEqual(self.model.nhist, len(self.hist0))
    self.assertTrue(np.allclose(self.model.init_hist(), self.hist0))

  def test_gradient(self):
    dfn = lambda x: self.model.q(x, self.T)
    ngrad = differentiate(dfn, self.hist_trial)
    grad = self.model.dq_da(self.hist_trial, self.T)
    self.assertTrue(np.allclose(ngrad, grad))

class TestLinearIsotropicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0

    self.hist0 = np.array([0.0])
    
    self.hist_trial = np.array([0.1])
    self.T = 300.0

    self.model = hardening.LinearIsotropicHardeningRule(self.s0, self.K)

  def test_properties(self):
    self.assertTrue(np.isclose(self.model.s0(self.T), self.s0))
    self.assertTrue(np.isclose(self.model.K(self.T), self.K))

  def test_relation(self):
    self.assertTrue(np.allclose(self.model.q(self.hist_trial, self.T), 
      np.array([-self.s0 - self.K * self.hist_trial[0]])))

class TestInterpolatedIsotropicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.points = [0.0, 0.01, 0.02, 0.05]
    self.values = [100.0, 110.0, 130.0, 135.0]
    self.ifn = interpolate.PiecewiseLinearInterpolate(self.points, self.values)
    self.hist0 = np.array([0.0])
    self.hist_trial = np.array([0.03])
    self.T = 300.0

    self.model = hardening.InterpolatedIsotropicHardeningRule(self.ifn)

  def test_relation(self):
    self.assertTrue(np.isclose(self.model.q(self.hist_trial, self.T)[0],
      -self.ifn(self.hist_trial[0])))


class TestVoceIsotropicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.s0 = 200.0
    self.R = 100.0
    self.d = 10.0

    self.hist0 = np.array([0.0])
    
    self.hist_trial = np.abs([0.25])
    self.T = 300.0

    self.model = hardening.VoceIsotropicHardeningRule(self.s0, self.R, self.d)

  def test_properties(self):
    self.assertTrue(np.isclose(self.model.s0(self.T), self.s0))
    self.assertTrue(np.isclose(self.model.R(self.T), self.R))
    self.assertTrue(np.isclose(self.model.d(self.T), self.d))

  def test_relation(self):
    self.assertTrue(np.allclose(self.model.q(self.hist_trial, self.T), 
      np.array([-self.s0 - self.R * (1 - np.exp(-self.d*self.hist_trial[0]))])))

class TestPowerLawIsotropicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.s0 = 200.0
    self.A = 100
    self.n = 0.2

    self.hist0 = np.array([0.0])
    
    self.hist_trial = np.abs([0.25])
    self.T = 300.0

    self.model = hardening.PowerLawIsotropicHardeningRule(self.s0, self.A, self.n)

  def test_relation(self):
    self.assertTrue(np.allclose(self.model.q(self.hist_trial, self.T), 
      np.array([-self.s0 - self.A * self.hist_trial[0]**self.n])))

class TestCombinedIsotropicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0

    self.hist0 = np.array([0.0])
    
    self.hist_trial = np.array([0.1])
    self.T = 300.0

    self.model1 = hardening.LinearIsotropicHardeningRule(self.s0, self.K)

    self.s0 = 200.0
    self.R = 100.0
    self.d = 10.0

    self.model2 = hardening.VoceIsotropicHardeningRule(self.s0, self.R, self.d)

    self.model = hardening.CombinedIsotropicHardeningRule([self.model1, self.model2])

  def test_combined(self):
    h1 = self.model1.q(self.hist_trial, self.T)
    h2 = self.model2.q(self.hist_trial, self.T)

    h = self.model.q(self.hist_trial, self.T)

    self.assertTrue(np.isclose(h, h1+h2))

class TestLinearKinematicHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.H = 1000.0

    self.hist0 = np.zeros((6,))
    
    self.hist_trial = np.array([50.0,25.0,75.0,10.0,15.0,50.0])
    self.hist_trial = self.hist_trial - np.array([1,1,1,0,0,0]) * sum(self.hist_trial[:3]) / 3.0
    self.T = 300.0

    self.model = hardening.LinearKinematicHardeningRule(self.H)

  def test_properties(self):
    self.assertTrue(np.isclose(self.model.H(self.T), self.H))

  def test_relation(self):
    self.assertTrue(np.allclose(self.model.q(self.hist_trial, self.T), 
      -self.hist_trial * self.H))

class TestCombinedHardening(unittest.TestCase, CommonHardening):
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0
    self.H = 1000.0

    self.hist0 = np.zeros((7,))
    self.hist_trial = np.array([50.0,75.0,100.0,25.0,10.0,15.0,60.0])
    self.hist_trial[1:] = make_dev(self.hist_trial[1:])
    self.T = 300.0

    self.iso = hardening.LinearIsotropicHardeningRule(self.s0, self.K)
    self.kin = hardening.LinearKinematicHardeningRule(self.H)

    self.model = hardening.CombinedHardeningRule(self.iso, self.kin)

  def test_relation(self):
    sb = np.zeros((7,))
    sb[0] = self.iso.q(self.hist_trial[0:1], self.T)
    sb[1:] = self.kin.q(self.hist_trial[1:], self.T)

    self.assertTrue(np.allclose(self.model.q(self.hist_trial, self.T), sb))

class CommonNonAssociative(object):
  """
    Common tests for non-associative hardening rules
  """
  def test_history(self):
    self.assertEqual(self.model.nhist, len(self.hist0))
    self.assertEqual(self.model.ninter, self.conform)
    self.assertTrue(np.allclose(self.model.init_hist(), self.hist0))

  def gen_stress(self):
    s = np.array([-150,200,-50,80,-20,30])
    return s

  def test_dq(self):
    a = self.gen_hist()

    dq_model = self.model.dq_da(a, self.T)
    dfn = lambda x: self.model.q(x, self.T)
    dq_num = differentiate(dfn, a)

    self.assertTrue(np.allclose(dq_model, dq_num))

  def test_dh_ds(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_ds(s, a, self.T)
    dfn = lambda x: self.model.h(x, a, self.T)
    dh_num = differentiate(dfn, s)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3, atol = 1.0e-5))

  def test_dh_da(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_da(s, a, self.T)
    dfn = lambda x: self.model.h(s, x, self.T)
    dh_num = differentiate(dfn, a)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3, atol = 1.0e-5))

  def test_dh_ds_time(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_ds_time(s, a, self.T)
    dfn = lambda x: self.model.h_time(x, a, self.T)
    dh_num = differentiate(dfn, s)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3))

  def test_dh_da_time(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_da_time(s, a, self.T)
    dfn = lambda x: self.model.h_time(s, x, self.T)
    dh_num = differentiate(dfn, a)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3, atol = 1.0e-4))

  def test_dh_ds_temp(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_ds_temp(s, a, self.T)
    dfn = lambda x: self.model.h_temp(x, a, self.T)
    dh_num = differentiate(dfn, s)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3))

  def test_dh_da_temp(self):
    s = self.gen_stress()
    a = self.gen_hist()

    dh_model = self.model.dh_da_temp(s, a, self.T)
    dfn = lambda x: self.model.h_temp(s, x, self.T)
    dh_num = differentiate(dfn, a)

    self.assertTrue(np.allclose(dh_model, dh_num, rtol = 1.0e-3))

class CommonGamma(object):
  def gen_alpha(self):
    return 0.15

  def test_dgamma(self):
    a = self.gen_alpha()
    should = self.model.dgamma(a, self.T)
    dfn = lambda a: self.model.gamma(a, self.T)
    num = differentiate(dfn, a)

    self.assertTrue(np.isclose(should, num))

class TestConstantGamma(unittest.TestCase, CommonGamma):
  def setUp(self):
    self.g = 100.0
    self.model = hardening.ConstantGamma(self.g)

    self.T = 300.0

  def test_properties(self):
    self.assertTrue(np.isclose(self.g, self.model.g(self.T)))

  def test_gamma(self):
    a = self.gen_alpha()
    self.assertTrue(np.isclose(self.g, self.model.gamma(a, self.T)))

class TestSatGamma(unittest.TestCase, CommonGamma):
  def setUp(self):
    self.g0 = 100.0
    self.gs = 200.0
    self.beta = 2.5

    self.T = 300.0

    self.model = hardening.SatGamma(self.gs, self.g0, self.beta)

  def test_properties(self):
    self.assertTrue(np.isclose(self.g0, self.model.g0(self.T)))
    self.assertTrue(np.isclose(self.gs, self.model.gs(self.T)))
    self.assertTrue(np.isclose(self.beta, self.model.beta(self.T)))

  def test_gamma(self):
    a = self.gen_alpha()
    should = self.gs + (self.g0 - self.gs)*np.exp(-self.beta * a)
    self.assertTrue(np.isclose(should, self.model.gamma(a, self.T)))

class TestChaboche(unittest.TestCase, CommonNonAssociative):
  """
    Chaboche model with arbitrary hardening
  """
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0

    self.n = 4
    self.cs = np.array(range(self.n)) * 10.0
    self.rs = np.array(range(1,self.n+1)) * 10.0
    self.gammas = [hardening.ConstantGamma(r) for r in self.rs]

    self.As = (np.array(range(self.n)) + 1.0)  * 10.0
    self.a_s = np.array([2.0]*self.n)

    self.iso = hardening.LinearIsotropicHardeningRule(self.s0, self.K)

    self.model = hardening.Chaboche(self.iso, list(self.cs), self.gammas,
        list(self.As), list(self.a_s))

    self.hist0 = np.zeros((1 + self.n*6,))
    self.conform = 7

    self.T = 300.0

  def gen_hist(self):
    hist =  np.array(range(1,self.n*6+2)) / (self.n*7)
    hist[1:] = (1.0 - 2.0 * hist[1:]) * 100.0
    for i in range(self.n):
      hist[1+i*6:1+(i+1)*6] = make_dev(hist[1+i*6:1+(i+1)*6])
    return hist

  def test_properties(self):
    self.assertEqual(self.n, self.model.n)
    self.assertTrue(np.allclose(self.model.c(self.T), self.cs))

  def test_q(self):
    h = self.gen_hist()
    q_model = self.model.q(h, self.T)
    q_exact = np.zeros((7,))
    q_exact[0] = self.iso.q(h[0:1], self.T)
    q_exact[1:] = sum(h[1+i*6:1+(i+1)*6] for i in range(self.n))
    self.assertTrue(np.allclose(q_model, q_exact))

  def test_h(self):
    alpha = self.gen_hist()
    s = self.gen_stress()
    sdev = make_dev(s)
    X = sum(alpha[1+i*6:1+(i+1)*6] for i in range(self.n))
    n = (sdev+X) / la.norm(sdev+X)

    h_model = self.model.h(s, alpha, self.T)

    h_exact = np.zeros((self.model.nhist,))
    h_exact[0] = np.sqrt(2.0/3.0)
    for i in range(self.n):
      h_exact[1+i*6:1+(i+1)*6] = -2.0 / 3.0 * self.cs[i] * n - np.sqrt(2.0/3.0)*self.gammas[i].gamma(alpha[0], self.T)*alpha[1+i*6:1+(i+1)*6]
    
    self.assertTrue(np.allclose(h_model, h_exact))

  def test_h_time(self):
    alpha = self.gen_hist()
    s = self.gen_stress()
    sdev = make_dev(s)
    X = sum(alpha[1+i*6:1+(i+1)*6] for i in range(self.n))
    n = (sdev+X) / la.norm(sdev+X)

    h_model = self.model.h_time(s, alpha, self.T)

    h_exact = np.zeros((self.model.nhist,))
    for i in range(self.n):
      Xi = alpha[1+i*6:1+(i+1)*6] 
      h_exact[1+i*6:1+(i+1)*6] = -self.As[i] * np.sqrt(3.0/2.0) * la.norm(Xi)**(self.a_s[i]-1.0) * Xi
    
    self.assertTrue(np.allclose(h_model, h_exact))

class TestChabocheNewFormLinear(unittest.TestCase, CommonNonAssociative):
  """
    Same as previous but with new-style setup
  """
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0

    self.n = 4
    self.cs = list(np.array(range(self.n)) / (1+self.n))
    self.rs = np.array(range(self.n)) * 10.0
    self.gammas = [hardening.ConstantGamma(r) for r in self.rs]
    self.As = [0.0] * self.n
    self.ns = [1.0] * self.n

    self.iso = hardening.LinearIsotropicHardeningRule(self.s0, self.K)

    self.model = hardening.Chaboche(self.iso, self.cs, self.gammas, 
        self.As, self.ns)

    self.hist0 = np.zeros((1 + self.n*6,))
    self.conform = 7

    self.T = 300.0

  def gen_hist(self):
    hist = np.array(range(1,2+self.n*6)) / (3*self.n)
    hist[1:] = (1.0 - 2.0 * hist[1:]) * 100.0
    for i in range(self.n):
      hist[1+i*6:1+(i+1)*6] = make_dev(hist[1+i*6:1+(i+1)*6])
    return hist

  def test_properties(self):
    self.assertEqual(self.n, self.model.n)
    self.assertTrue(np.allclose(self.model.c(self.T), self.cs))

  def test_q(self):
    h = self.gen_hist()
    q_model = self.model.q(h, self.T)
    q_exact = np.zeros((7,))
    q_exact[0] = self.iso.q(h[0:1], self.T)
    q_exact[1:] = sum(h[1+i*6:1+(i+1)*6] for i in range(self.n))
    self.assertTrue(np.allclose(q_model, q_exact))

  def test_h(self):
    alpha = self.gen_hist()
    s = self.gen_stress()
    sdev = make_dev(s)
    X = sum(alpha[1+i*6:1+(i+1)*6] for i in range(self.n))
    n = (sdev+X) / la.norm(sdev+X)

    h_model = self.model.h(s, alpha, self.T)

    h_exact = np.zeros((self.model.nhist,))
    h_exact[0] = np.sqrt(2.0/3.0)
    for i in range(self.n):
      h_exact[1+i*6:1+(i+1)*6] = -2.0 / 3.0 * self.cs[i] * n - np.sqrt(2.0/3.0)*self.gammas[i].gamma(alpha[0], self.T)*alpha[1+i*6:1+(i+1)*6]
    
    self.assertTrue(np.allclose(h_model, h_exact))

class TestChabocheNewFormSat(unittest.TestCase, CommonNonAssociative):
  """
    Nonlinear saturation function
  """
  def setUp(self):
    self.s0 = 200.0
    self.K = 1000.0

    self.n = 2
    self.cs = list(np.array(range(self.n)) / (2*self.n)+10.0)
    self.g0s = np.array(range(1,self.n+1)) * 10.0
    self.gss = 2.0 * self.g0s
    self.betas = np.array(range(self.n)) / self.n
    self.gammas = [hardening.SatGamma(a, b, c) for a,b,c in zip(self.g0s, self.gss, self.betas)]
    self.As = [0.0] * self.n
    self.ns = [1.0] * self.n

    self.iso = hardening.LinearIsotropicHardeningRule(self.s0, self.K)

    self.model = hardening.Chaboche(self.iso, self.cs, self.gammas, self.As,
        self.ns)

    self.hist0 = np.zeros((1 + self.n*6,))
    self.conform = 7

    self.T = 300.0

  def gen_hist(self):
    hist = np.array(range(1, 2 + self.n*6))
    hist[1:] = (1.0 - 2.0 * hist[1:]) * 100.0
    for i in range(self.n):
      hist[1+i*6:1+(i+1)*6] = make_dev(hist[1+i*6:1+(i+1)*6])
    return hist

  def test_properties(self):
    self.assertEqual(self.n, self.model.n)
    self.assertTrue(np.allclose(self.model.c(self.T), self.cs))

  def test_q(self):
    h = self.gen_hist()
    q_model = self.model.q(h, self.T)
    q_exact = np.zeros((7,))
    q_exact[0] = self.iso.q(h[0:1], self.T)
    q_exact[1:] = sum(h[1+i*6:1+(i+1)*6] for i in range(self.n))
    self.assertTrue(np.allclose(q_model, q_exact))

  def test_h(self):
    alpha = self.gen_hist()
    s = self.gen_stress()
    sdev = make_dev(s)
    X = sum(alpha[1+i*6:1+(i+1)*6] for i in range(self.n))
    n = (sdev+X) / la.norm(sdev+X)

    h_model = self.model.h(s, alpha, self.T)

    h_exact = np.zeros((self.model.nhist,))
    h_exact[0] = np.sqrt(2.0/3.0)
    for i in range(self.n):
      h_exact[1+i*6:1+(i+1)*6] = -2.0 / 3.0 * self.cs[i] * n - np.sqrt(2.0/3.0)*self.gammas[i].gamma(alpha[0], self.T)*alpha[1+i*6:1+(i+1)*6]
    
    self.assertTrue(np.allclose(h_model, h_exact))

class TestChabocheTempTerm(unittest.TestCase, CommonNonAssociative):
  """
    Same as previous but with new-style setup
  """
  def setUp(self):
    s0 = 200.0
    K = 1000.0

    self.n = 2
    cvals1 = [1000.0,10.0]
    cvals2 = [100.0,100.0]
    Ts = [0.0,1000.0]

    cs = [interpolate.PiecewiseLinearInterpolate(Ts, [ci,cj]) for ci,cj in zip(cvals1, cvals2)]
    rs = [1.0e-2,1.0]
    gammas = [hardening.ConstantGamma(r) for r in rs]
    self.As = [0.0] * self.n
    self.ns = [1.0] * self.n

    iso = hardening.LinearIsotropicHardeningRule(s0, K)

    self.model = hardening.Chaboche(iso, cs, gammas, self.As, self.ns)

    self.hist0 = np.zeros((1 + self.n*6,))
    self.conform = 7

    self.T = 300.0
  
  def gen_hist(self):
    hist = np.array(range(1,2+self.n*6)) / (3*self.n)
    hist[1:] = (1.0 - 2.0 * hist[1:]) * 100.0
    for i in range(self.n):
      hist[1+i*6:1+(i+1)*6] = make_dev(hist[1+i*6:1+(i+1)*6])
    return hist



#!/usr/bin/env python3

from neml import history, interpolate
from neml.math import tensors, rotations, matrix
from neml.cp import crystallography, slipharden, sliprules

from common import differentiate
from nicediff import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonSlipRule(object):
  def test_d_slip_d_stress(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        d = self.model.d_slip_d_s(g, i, self.S, self.Q, self.H, self.L, self.T, self.fixed)
        nd = diff_scalar_symmetric(lambda s: self.model.slip(g, i, s, self.Q, self.H, 
          self.L, self.T, self.fixed), self.S)
        self.assertEqual(d, nd)

  def test_d_slip_d_hist(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        d = np.array(self.model.d_slip_d_h(g, i, self.S, self.Q, self.H, self.L, self.T, self.fixed))
        nd = np.array(diff_history_scalar(lambda h: self.model.slip(g, i, self.S, self.Q, h,
          self.L, self.T, self.fixed), self.H))
        self.assertTrue(np.allclose(nd.reshape(d.shape), d))

  def test_d_hist_rate_d_stress(self):
    d = np.array(self.model.d_hist_rate_d_stress(self.S, self.Q, self.H, self.L, self.T, self.fixed))
    nd = diff_history_symmetric(lambda s: self.model.hist_rate(s, self.Q, self.H, self.L,
      self.T, self.fixed), self.S)
    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

  def test_d_hist_rate_d_hist(self):
    d = np.array(self.model.d_hist_rate_d_hist(self.S, self.Q, self.H, self.L, self.T, self.fixed))
    nd = diff_history_history(lambda h: self.model.hist_rate(self.S, self.Q, h, self.L,
      self.T, self.fixed), self.H)
    d = d.reshape(nd.shape)
    print(d)
    print(nd)
    np.set_printoptions(precision = 3, suppress = True)
    print(d[:12,:12])
    print(nd[:12,:12])
    self.assertTrue(np.allclose(nd.reshape(d.shape), d))

class CommonSlipMultiStrengthSlipRule(object):
  def test_setup_history(self):
    model_hist = history.History()
    self.model.populate_history(model_hist)
    self.model.init_history(model_hist)
    
    manual_hist = history.History()
    for strength in self.strengths:
      Hs = history.History()
      strength.populate_history(Hs)
      strength.init_history(Hs)
      manual_hist.add_union(Hs)

    self.assertTrue(np.allclose(np.array(model_hist),
      np.array(manual_hist)))

  def test_history_rate(self):
    reference = self.model.hist_rate(self.S, self.Q, self.H, self.L, self.T,
        self.fixed)
    combined = history.History()
    for strength in self.strengths:
      combined.add_union(strength.hist(self.S, self.Q, self.H, self.L, self.T, self.model, self.fixed))

    self.assertTrue(np.allclose(reference, combined))

  def test_slip_rate(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        rs = self.L.shear(g, i, self.Q, self.S)
        self.assertTrue(np.isclose(self.model.slip(g, i, self.S, self.Q, self.H, self.L, self.T, self.fixed),
          self.model.sslip(g, i, rs, self.strength_values, self.T)))

  def test_slip_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        rs = self.L.shear(g, i, self.Q, self.S)
        nd = differentiate(lambda t: self.model.sslip(g, i, t, self.strength_values, self.T),
            rs)
        d = self.model.d_sslip_dtau(g, i, rs, self.strength_values, self.T)

        self.assertTrue(np.isclose(d, nd, rtol = 1e-4))

  def test_slip_strength(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        rs = self.L.shear(g, i, self.Q, self.S)

        nd = differentiate(lambda s: self.model.sslip(g, i, rs, s, self.T), np.array(self.strength_values))
        d = self.model.d_sslip_dstrength(g, i, rs, self.strength_values, self.T)

        self.assertTrue(np.allclose(nd,d))

class TestKinematicPowerLawSlip(unittest.TestCase, CommonSlipRule, CommonSlipMultiStrengthSlipRule):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength0 = 15.0
    self.strength1 = 20.0
    self.strength2 = 20.0
    self.H = history.History()
    self.H.add_scalar("strength_#0")
    self.H.add_scalar("strength_#1")
    self.H.add_scalar("strength_#2")
    self.H.set_scalar("strength_#0", self.strength0)
    self.H.set_scalar("strength_#1", self.strength1)
    self.H.set_scalar("strength_#2", self.strength2)

    self.T = 300.0

    self.tau0_1 = 10.0
    self.tau_sat_1 = 50.0
    self.b_1 = 2.5

    self.tau0_2 = 15.0
    self.tau_sat_2 = 55.0
    self.b_2 = 3.0

    self.tau0_3 = 10.0
    self.tau_sat_3 = 45.0
    self.b_3 = 5.0

    self.backstrength = slipharden.VoceSlipHardening(self.tau_sat_1, self.b_1, self.tau0_1)
    self.isostrength = slipharden.VoceSlipHardening(self.tau_sat_2, self.b_2, self.tau0_2)
    self.flowresistance = slipharden.VoceSlipHardening(self.tau_sat_3, self.b_3, self.tau0_3)
    self.strengths = [self.backstrength, self.isostrength, self.flowresistance]

    self.strength_values = [self.strength0 + self.tau0_1, self.strength1 + self.tau0_2, self.strength2 + self.tau0_3]
    
    self.g0 = 1.0
    self.n = 3.0
    self.model = sliprules.KinematicPowerLawSlipRule(self.backstrength, self.isostrength, self.flowresistance, self.g0, self.n)

    self.tau = 80.0

    self.fixed = history.History()

  def test_sslip(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        bs = self.strength_values[0]
        iso = self.strength_values[1]
        fr = self.strength_values[2]
        inv = (np.abs(self.tau - bs) - iso) / fr

        expected = self.g0 * np.abs(inv) ** (self.n) * np.sign(self.tau - bs)

        actual = self.model.sslip(g, i, self.tau, self.strength_values, self.T)

        self.assertTrue(np.isclose(expected, actual))

class TestKinematicPowerLawSlipComplicated(unittest.TestCase, CommonSlipRule):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))*3

    self.H = history.History()

    self.T = 300.0

    E = 160000.0
    nu = 0.31

    self.g0 = 1.0e-4
    self.n = 10.0

    K = E / 50.0
    s0 = 75.0

    M = matrix.SquareMatrix(self.L.ntotal, type = "dense", 
        data = [K] * (self.L.ntotal * self.L.ntotal))

    self.isostrength = slipharden.GeneralLinearHardening(M, [s0/2] * self.L.ntotal, 
        absval = False)
    self.flowresistance = slipharden.GeneralLinearHardening(M, [s0/2] * self.L.ntotal, 
        absval = False)
    self.backstrength = slipharden.GeneralLinearHardening(M, [0] * self.L.ntotal, 
        absval = False)

    self.strengths = [self.backstrength, self.isostrength, self.flowresistance]

    self.model = sliprules.KinematicPowerLawSlipRule(self.backstrength, self.isostrength, self.flowresistance, self.g0, self.n)

    self.model.populate_history(self.H)
    self.model.init_history(self.H)

    self.strength_values = np.linspace(0,10,36) + 5.0
    self.H.copy_data(self.strength_values)

    self.fixed = history.History()

class CommonSlipStrengthSlipRule(CommonSlipMultiStrengthSlipRule):
  def test_init_hist(self):
    H1 = history.History()
    self.model.populate_history(H1)
    self.model.init_history(H1)

    H2 = history.History()
    self.strengthmodel.populate_history(H2)
    self.strengthmodel.init_history(H2)

    self.assertTrue(np.allclose(np.array(H1),
      np.array(H2)))

  def test_slip(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        rs = self.L.shear(g, i, self.Q, self.S)
        strength = self.strength + self.static
        self.assertTrue(np.isclose(self.model.slip(g, i, self.S, self.Q, self.H, self.L, self.T, self.fixed),
          self.model.scalar_sslip(g, i, rs, strength, self.T)))

  def test_d_hist_rate(self):
    self.assertTrue(np.allclose(
      np.array(self.model.hist_rate(self.S, self.Q, self.H, self.L, self.T, self.fixed)),
      np.array(self.strengthmodel.hist(self.S, self.Q, self.H, self.L, self.T, self.model, self.fixed))))

  def test_d_sslip_d_tau(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        nd = differentiate(lambda t: self.model.scalar_sslip(g, i, t, self.strength, self.T),
            self.tau)
        d = self.model.scalar_d_sslip_dtau(g, i, self.tau, self.strength, self.T)
        self.assertTrue(np.isclose(nd,d))

  def test_d_sslip_d_strength(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        nd = differentiate(lambda s: self.model.scalar_sslip(g, i, self.tau, s, self.T), self.strength)
        d = self.model.scalar_d_sslip_dstrength(g, i, self.tau, self.strength, self.T)
        self.assertTrue(np.isclose(nd, d))

class TestPowerLawSlip(unittest.TestCase, CommonSlipStrengthSlipRule, CommonSlipRule):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength = 35.0
    self.H = history.History()
    self.H.add_scalar("strength")
    self.H.set_scalar("strength", self.strength)

    self.T = 300.0

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0)
    self.strengths = [self.strengthmodel]

    self.static = self.tau0

    self.strength_values = [self.strength + self.static]
    
    self.g0 = 1.0
    self.n = 3.0
    self.model = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.tau = 33.0

    self.fixed = history.History()

  def test_scalar_rate(self):
    for g in range(self.L.ngroup):
      for i in range(self.L.nslip(g)):
        self.assertTrue(np.isclose(self.model.scalar_sslip(g, i, self.tau, self.strength, self.T),
          self.g0 * np.abs(self.tau/self.strength)**(self.n-1.0) * self.tau/self.strength))


class TestBiVoceSlip(unittest.TestCase, CommonSlipStrengthSlipRule, CommonSlipRule):
  def setUp(self):
    self.L = crystallography.CubicLattice(1.0)
    self.L.add_slip_system([1,1,0],[1,1,1])
    
    self.Q = rotations.Orientation(35.0,17.0,14.0, angle_type = "degrees")
    self.S = tensors.Symmetric(np.array([
      [100.0,-25.0,10.0],
      [-25.0,-17.0,15.0],
      [10.0,  15.0,35.0]]))
    self.strength_1 = 35.0
    self.strength_2 = 25.0
    self.strength = self.strength_1 + self.strength_2
    self.H = history.History()
    self.H.add_scalar("strength0")
    self.H.set_scalar("strength0", self.strength_1)
    self.H.add_scalar("strength1")
    self.H.set_scalar("strength1", self.strength_2)

    self.T = 300.0

    self.tau0 = 10.0
    self.tau_sat = 50.0
    self.b = 2.5

    self.strengthmodel = slipharden.SumSlipSingleStrengthHardening(
        [slipharden.VoceSlipHardening(self.tau_sat, self.b, self.tau0),
          slipharden.VoceSlipHardening(self.tau_sat/2, self.b/2, self.tau0/2)])
    self.static = self.tau0 + self.tau0 / 2

    self.strengths = [self.strengthmodel]
    self.strength_values = [self.strength + self.static]
    
    self.g0 = 1.0
    self.n = 3.0
    self.model = sliprules.PowerLawSlipRule(self.strengthmodel, self.g0, self.n)

    self.tau = 33.0

    self.fixed = history.History()



import sys
sys.path.append('..')

from neml import general_flow, visco_flow, elasticity, surfaces, hardening
from common import *

import unittest
import numpy as np
import numpy.linalg as la

class CommonGeneralFlow(object):
  def test_ds_ds(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.s(x, h_np1, e_dot, T_np1, T_dot)
    num = differentiate(dfn, s_np1)
    should = self.model.ds_ds(s_np1, h_np1, e_dot, T_np1, T_dot)
    
    self.assertTrue(np.allclose(num, should))

  def test_ds_da(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.s(s_np1, x, e_dot, T_np1, T_dot)
    num = differentiate(dfn, h_np1)
    should = self.model.ds_da(s_np1, h_np1, e_dot, T_np1, T_dot)
    
    self.assertTrue(np.allclose(num, should, rtol = 1.0e-3))

  def test_ds_de(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.s(s_np1, h_np1, x, T_np1, T_dot)
    num = differentiate(dfn, e_dot, eps = 1.0e-3)
    should = self.model.ds_de(s_np1, h_np1, e_dot, T_np1, T_dot)

    self.assertTrue(np.allclose(num, should, rtol = 1.0e-3))

  def test_da_ds(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.a(x, h_np1, e_dot, T_np1, T_dot)
    num = differentiate(dfn, s_np1)
    should = self.model.da_ds(s_np1, h_np1, e_dot, T_np1, T_dot)
    
    self.assertTrue(np.allclose(num, should, rtol = 1.0e-3))

  def test_da_da(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.a(s_np1, x, e_dot, T_np1, T_dot)
    num = differentiate(dfn, h_np1)
    should = self.model.da_da(s_np1, h_np1, e_dot, T_np1, T_dot)

    self.assertTrue(np.allclose(num, should, rtol = 1.0e-3))

  def test_da_de(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    dfn = lambda x: self.model.a(s_np1, h_np1, x, T_np1, T_dot)
    num = differentiate(dfn, e_dot, eps = 1.0e-3)
    should = self.model.da_de(s_np1, h_np1, e_dot, T_np1, T_dot)

    self.assertTrue(np.allclose(num, should, rtol = 1.0e-3))


class CommonTVPFlow(object):
  def test_history(self):
    self.assertEqual(len(self.h_n), self.model.nhist)
    self.assertTrue(np.allclose(self.h_n, self.model.init_hist()))

  def test_srate(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    should = np.dot(self.emodel.C(T_np1), 
        e_dot - self.vmodel.g(s_np1, h_np1, T_np1) * 
        self.vmodel.y(s_np1, h_np1, T_np1) - 
        self.vmodel.g_temp(s_np1, h_np1, T_np1) * T_dot -
        self.vmodel.g_time(s_np1, h_np1, T_np1))

    actual = self.model.s(s_np1, h_np1, e_dot, T_np1, T_dot)

    self.assertTrue(np.allclose(should, actual))

  def test_hrate(self):
    t_np1 = self.gen_t()
    e_np1 = self.gen_e()
    e_dot = self.gen_edot(e_np1, t_np1)
    T_np1 = self.gen_T()
    T_dot = self.gen_Tdot(T_np1, t_np1)
    s_np1 = self.gen_stress()
    h_np1 = self.gen_hist()

    should = (self.vmodel.h(s_np1, h_np1, T_np1) * 
      self.vmodel.y(s_np1, h_np1, T_np1) + 
      self.vmodel.h_temp(s_np1, h_np1, T_np1) * T_dot +
      self.vmodel.h_time(s_np1, h_np1, T_np1))

    actual = self.model.a(s_np1, h_np1, e_dot, T_np1, T_dot)

    self.assertTrue(np.allclose(should, actual))


class TestTVPCheboche(unittest.TestCase, CommonGeneralFlow, CommonTVPFlow):
  def setUp(self):
    n = 20.0
    eta = 108.0
    sY = 89.0

    Q = 165.0
    b = 12.0
    
    self.m = 3

    C1 = 80.0e3
    C2 = 14.02e3
    C3 = 3.333e3

    y1 = 0.9e3
    y2 = 1.5e3
    y3 = 1.0

    surface = surfaces.IsoKinJ2()
    iso = hardening.VoceIsotropicHardeningRule(sY, Q, b)
    cs = [C1, C2, C3]
    gs = [y1, y2, y3]
    As = [0.0, 0.0, 0.0]
    ns = [1.0, 1.0, 1.0]
    gmodels = [hardening.ConstantGamma(g) for g in gs]
    hmodel = hardening.Chaboche(iso, cs, gmodels, As, ns)

    fluidity = visco_flow.ConstantFluidity(eta)

    self.vmodel = visco_flow.ChabocheFlowRule(
        surface, hmodel, fluidity, n)

    E = 92000.0
    nu = 0.3

    mu = E/(2*(1+nu))
    K = E/(3*(1-2*nu))

    self.emodel = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    self.model = general_flow.TVPFlowRule(self.emodel, self.vmodel)

    self.T_n = 300.0
    self.e_n = np.zeros((6,))
    self.t_n = 0.0
    self.h_n = np.zeros((1+6*self.m,))

  def gen_hist(self):
    h = np.array([0.1,50,-30,40,60,-80,-30,-100,-15,-30,-50,100,50,60,-60,50,
      30,90,40])
    for i in range(self.m):
      h[1+i*6:1+(i+1)*6] = make_dev(h[1+i*6:1+(i+1)*6])
    
    return h

  def gen_stress(self):
    return np.array([200.0,-200.0,100.0,50.0,25.0,-50.0])

  def gen_e(self):
    return np.array([0.025,-0.01,-0.02,0.01,0.02,-0.03])

  def gen_T(self):
    return 350

  def gen_t(self):
    return 1.25

  def gen_dt(self, t):
    return t - self.t_n

  def gen_edot(self, e, t):
    return (e - self.e_n) / self.gen_dt(t)

  def gen_Tdot(self, T, t):
    return (T - self.T_n) / self.gen_dt(t)


class TestTVPYaguchi(unittest.TestCase, CommonGeneralFlow, CommonTVPFlow):
  def setUp(self):
    self.vmodel = visco_flow.YaguchiGr91FlowRule()

    E = 92000.0
    nu = 0.3

    mu = E/(2*(1+nu))
    K = E/(3*(1-2*nu))

    self.emodel = elasticity.IsotropicLinearElasticModel(mu, "shear",
        K, "bulk")

    self.model = general_flow.TVPFlowRule(self.emodel, self.vmodel)

    self.T_n = 300.0
    self.e_n = np.zeros((6,))
    self.t_n = 0.0
    self.h_n = np.zeros((2+12,))

  def gen_hist(self):
    h = np.array([50,-30,40,60,-80,-30,-100,-15,-30,-50,100,50,5,2.5])
    for i in range(2):
      h[1+i*6:1+(i+1)*6] = make_dev(h[1+i*6:1+(i+1)*6])
    
    return h

  def gen_stress(self):
    return np.array([200.0,-200.0,100.0,50.0,25.0,-50.0])

  def gen_e(self):
    return np.array([0.025,-0.01,-0.02,0.01,0.02,-0.03])

  def gen_T(self):
    return 350

  def gen_t(self):
    return 1.25

  def gen_dt(self, t):
    return t - self.t_n

  def gen_edot(self, e, t):
    return (e - self.e_n) / self.gen_dt(t)

  def gen_Tdot(self, T, t):
    return (T - self.T_n) / self.gen_dt(t)
